import logging
from abc import ABC
from typing import List

import openai

logger = logging.getLogger(__name__)


class Embed(ABC):
    def __init__(
        self,
        api_key: str,
        api_base_url: str,
        model_name: str,
        org_id: str,
        **model_kwargs,
    ):
        """
        Initialize the Embed class with OpenAI API configurations.

        Args:
            api_key (str): The API key for OpenAI.
            api_base_url (str): The base URL for the OpenAI API.
            model_name (str): The name of the embedding model.
            org_id (str): The organization ID for OpenAI.
            **model_kwargs: Additional keyword arguments for the model.
        """
        self.api_key = api_key
        self.api_base_url = api_base_url
        self.model_name = model_name
        self.org_id = org_id
        self.model_kwargs = model_kwargs

        self.client = openai.OpenAI(
            api_key=self.api_key, base_url=self.api_base_url, organization=self.org_id
        )

    def embed(self, texts: List[str]) -> List[List[float]]:
        """
        Generate embeddings for a list of texts using the OpenAI API.

        Args:
            texts (List[str]): A list of text strings to generate embeddings for.

        Returns:
            List[List[float]]: A list of embeddings, where each embedding is a list of floats.
        """
        response = self.client.embeddings.create(
            input=texts,
            model=self.model_name,
            **self.model_kwargs,
        )
        return [embedding.embedding for embedding in response.data]


if __name__ == "__main__":

    import argparse

    from cicada.core.utils import colorstring, load_config, setup_logging

    parser = argparse.ArgumentParser(description="Embedding Model")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    args = parser.parse_args()

    setup_logging()

    embed_config = load_config(args.config, "embed")

    embed = Embed(
        embed_config["api_key"],
        embed_config.get("api_base_url"),
        embed_config.get("model_name", "text-embedding-3-small"),
        embed_config.get("org_id"),
        **embed_config.get("model_kwargs", {}),
    )

    texts = ["This is a test document.", "Another test document."]
    embeddings = embed.embed(texts)
    logger.info(colorstring(f"Generated embeddings: {embeddings}", "white"))



import logging
from typing import List

from cicada.core import model
from cicada.core.basics import PromptBuilder
from cicada.core.utils import get_image_paths, image_to_base64

logger = logging.getLogger(__name__)


def load_images(path: str) -> List[str]:
    """
    Load images from a folder and convert them to base64.

    Args:
        path (str): Path to the folder containing images.

    Returns:
        List[bytes]: List of images in base64 format.
    """
    if path:
        return [image_to_base64(img) for img in get_image_paths(path)]
    else:
        return []


class VisualQA(model.MultiModalModel):
    def __init__(
        self,
        api_key,
        api_base_url,
        model_name,
        org_id,
        prompt_templates,
        **model_kwargs,
    ):
        """
        Initialize the VisualQA model.
        Args:
            api_key (str): API key for the model.
            api_base_url (str): Base URL for the API.
            model_name (str): Name of the model.
            org_id (str): Organization ID.
            prompt_templates (dict): Dictionary of prompt templates.
            **model_kwargs: Additional keyword arguments for the model.
        """
        super().__init__(
            api_key,
            api_base_url,
            model_name,
            org_id,
            **model_kwargs,
        )
        self.visual_qa_prompts = prompt_templates

    def generate_questions(
        self,
        design_goal: str,
        reference_images: List[str] | None,
        num_questions: int = 5,
    ) -> List[str]:
        """
        Generate questions based on the design goal and images.
        Args:
            design_goal (str): Text description of the design goal.
            reference_images (List[str]): List of reference images paths.
            num_questions (int): Number of questions to generate.

        Returns:
            List[str]: List of generated questions.
        """
        prompt = self.visual_qa_prompts["question_generation"][
            "prompt_template"
        ].format(design_goal=design_goal, num_questions=num_questions)

        pb = PromptBuilder()
        pb.add_system_message(
            self.visual_qa_prompts["question_generation"]["system_prompt"]
        )
        pb.add_user_message(prompt)
        if reference_images:
            pb.add_text("The following is a set of reference images:")
            pb.add_images(reference_images)

        response = self.query(prompt_builder=pb, stream=self.stream)["content"]

        questions = [q.strip() for q in response.split("\n") if q.strip()]
        return questions

    def generate_answers(
        self,
        design_goal: str,
        questions: List[str],
        reference_images: List[str] | None,
        rendered_images: List[str],
    ) -> dict:
        """
        Generate answers to the provided questions based on the design goal and images.

        Args:
            design_goal (str): Text description of the design goal.
            questions (List[str]): List of questions to answer.
            reference_images (List[str]): List of reference images in str.
            rendered_images (List[str]): List of rendered images in str.

        Returns:
            dict: Dictionary of questions and their corresponding answers.
        """
        questions_text = "\n".join([f"{i+1}. {q}" for i, q in enumerate(questions)])

        prompt = self.visual_qa_prompts["answer_generation"]["prompt_template"].format(
            design_goal=design_goal, questions=questions_text
        )

        # construct using PromptBuilder
        pb = PromptBuilder()
        pb.add_system_message(
            self.visual_qa_prompts["answer_generation"]["system_prompt"]
        )
        pb.add_user_message(prompt)  # Add the formatted prompt to the PromptBuilder
        if reference_images:
            pb.add_text("The followings are reference images of design goal:")
            pb.add_images(reference_images)
        pb.add_text("The following is a set of rendered object images:")
        pb.add_images(rendered_images)

        response = self.query(prompt_builder=pb, stream=self.stream)["content"]

        return response

    def automated_qa(
        self,
        design_goal: str,
        reference_images: List[str] | None,
        rendered_images: List[str],
        num_questions: int = 5,
    ) -> dict:
        """
        Generate questions and then generate answers to those questions.

        Args:
            design_goal (str): Text description of the design goal.
            reference_images (List[str]): List of reference images in str.
            rendered_images (List[str]): List of rendered images in str.
            num_questions (int): Number of questions to generate.

        Returns:
            dict: Dictionary containing the generated questions and answers.
        """
        questions = self.generate_questions(
            design_goal=design_goal,
            reference_images=reference_images,
            num_questions=num_questions,
        )
        answers = self.generate_answers(
            design_goal, questions, reference_images, rendered_images
        )
        return {"questions": questions, "answers": answers}


if __name__ == "__main__":
    import argparse
    import json

    from cicada.core.utils import (
        colorstring,
        cprint,
        get_image_paths,
        image_to_base64,
        load_config,
        load_prompts,
        parse_design_goal,
        setup_logging,
    )

    setup_logging()

    parser = argparse.ArgumentParser(description="Visual QA Model")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    parser.add_argument(
        "--prompts", default="prompts", help="Path to the prompts YAML file or folder"
    )
    parser.add_argument(
        "--mode",
        required=True,
        choices=["question_generation", "answer_generation", "automated_qa"],
        help="Operating mode: question_generation, answer_generation, or automated_qa",
    )
    parser.add_argument(
        "--design_goal",
        help="Text description of the design goal or path to a JSON file containing the design goal",
    )
    parser.add_argument("--questions", nargs="+", help="List of questions to answer")
    parser.add_argument(
        "--reference_images", help="Path to the folder containing reference images"
    )
    parser.add_argument(
        "--rendered_images", help="Path to the folder containing rendered object images"
    )
    parser.add_argument(
        "--num_questions", type=int, default=5, help="Number of questions to generate"
    )
    args = vars(parser.parse_args())

    config = load_config(args["config"], "visual_qa")
    prompt_templates = load_prompts(args["prompts"], "visual_qa")

    visual_qa = VisualQA(
        config["api_key"],
        config.get("api_base_url"),
        config.get("model_name", "gpt-4"),
        config.get("org_id"),
        prompt_templates,
        **config.get("model_kwargs", {}),
    )

    if args["mode"] == "question_generation":
        if not args["design_goal"]:
            raise ValueError("Design goal is required for question generation.")
        design_goal = parse_design_goal(args["design_goal"])
        questions = visual_qa.generate_questions(
            design_goal=design_goal,
            reference_images=args["reference_images"],
            num_questions=args["num_questions"],
        )
        cprint(json.dumps(questions, indent=4), "cyan")

    elif args["mode"] == "answer_generation":
        if not args["questions"]:
            raise ValueError("Questions are required for answer generation.")
        questions = args["questions"]
        design_goal = (
            parse_design_goal(args["design_goal"]) if args["design_goal"] else ""
        )
        answers = visual_qa.generate_answers(
            design_goal=design_goal,
            questions=questions,
            reference_images=args["reference_images"],
            rendered_images=args["rendered_images"],
        )
        cprint(json.dumps(answers, indent=4), "cyan")

    elif args["mode"] == "automated_qa":
        if not args["design_goal"]:
            raise ValueError("Design goal is required for automated QA.")
        design_goal = parse_design_goal(args["design_goal"])
        qa_result = visual_qa.automated_qa(
            design_goal=design_goal,
            reference_images=args["reference_images"],
            rendered_images=args["rendered_images"],
            num_questions=args["num_questions"],
        )
        cprint(json.dumps(qa_result, indent=4), "cyan")

    else:
        raise ValueError("Invalid mode selected.")



# src/cicada/__init__.py
from .core import *
from .retrieval import *

# Conditional import for codecad features
try:
    from .geometry_pipeline import *
    from .workflow import *
    from .describe import *
    from .coding import *
    from .feedback import *
except ImportError:
    pass

__version__ = "0.7.0"



import argparse
import logging
import os
import sys
from typing import List

from tqdm import tqdm

_current_dir = os.path.dirname(os.path.abspath(__file__))
_parent_dir = os.path.dirname(_current_dir)
sys.path.extend([_current_dir, _parent_dir])

from common import llm
from common.utils import load_config, load_prompts
from describe.utils import save_descriptions

# Configure logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


class AssistLLM(llm.LanguageModel):
    def __init__(
        self,
        api_key,
        api_base_url,
        model_name,
        org_id,
        prompt_templates,
        **model_kwargs,
    ):
        super().__init__(
            api_key,
            api_base_url,
            model_name,
            org_id,
            **model_kwargs,
        )
        self.user_prompt_templates = prompt_templates.get("user_prompt_template", {})
        self.system_prompt_template = prompt_templates.get("system_prompt_template", "")

    def distill_what_it_is(self, input_text: str) -> str:
        user_prompt_template = self.user_prompt_templates.get("what_is_this_object")
        prompt = user_prompt_template.format(generated_description=input_text)

        relevant_text = self.query(prompt)
        return relevant_text

    def extract_object_description(self, input_text: str) -> str:
        user_pompt_template = self.user_prompt_templates.get(
            "object_description_with_features"
        )
        prompt = user_pompt_template.format(generated_description=input_text)

        relevant_text = self.query(prompt, system_prompt=self.system_prompt_template)
        return relevant_text

    def extract_parts_list(self, input_text: str) -> str:
        user_prompt_template = self.user_prompt_templates.get("potential_parts_list")
        prompt = user_prompt_template.format(generated_description=input_text)

        relevant_text = self.query(prompt, system_prompt=self.system_prompt_template)
        return relevant_text

    def extract_building_steps(self, input_text: str) -> str:
        user_prompt_template = self.user_prompt_templates.get(
            "CAD_construction_instructions"
        )
        prompt = user_prompt_template.format(generated_description=input_text)

        relevant_text = self.query(prompt, system_prompt=self.system_prompt_template)
        return relevant_text


def load_object_metadata(task_path: str) -> List[dict]:
    """
    param task_path: Path to the task YAML file or path to certain metadata.json or directory containing a metadata.json

    returns: List of metadata dictionaries
    """
    if os.path.isdir(task_path):
        # single metadata
        metadata = {
            "base_path": task_path,
            "metadata": load_config(os.path.join(task_path, "metadata.json")),
        }
        metadata_collection = [metadata]
    elif os.path.isfile(task_path) and task_path.endswith("metadata.json"):
        # single metadata
        metadata = {
            "base_path": os.path.dirname(task_path),
            "metadata": load_config(task_path),
        }
        metadata_collection = [metadata]
    elif os.path.isfile(task_path) and task_path.endswith("tasks.yaml"):
        tasks = load_config(task_path)
        metadata_collection = [
            {
                "base_path": obj["base_path"],
                "metadata": load_config(
                    os.path.join(obj["base_path"], "metadata.json")
                ),
            }
            for obj in tasks
        ]
    return metadata_collection


def _main():
    parser = argparse.ArgumentParser(description="Assistive Large Language Model")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    parser.add_argument(
        "--prompts", default="prompts.yaml", help="Path to the prompts YAML file"
    )
    parser.add_argument(
        "-t",
        "--task",
        default="tasks.yaml",
        help="Path to the task YAML file or directory containing images from a single object",
    )
    parser.add_argument(
        "--save",
        action="store_true",
        help="Save generated descriptions to metadata.yaml",
    )
    args = parser.parse_args()

    config = load_config(args.config)
    tasks = load_object_metadata(args.task)

    assist_llm_config = config["assist-llm"]

    llm = AssistLLM(
        assist_llm_config["api_key"],
        assist_llm_config.get("api_base_url"),
        assist_llm_config.get("model_name", "gpt-4o-mini"),
        assist_llm_config.get("org_id"),
        load_prompts(args.prompts, "assist-llm"),
        **assist_llm_config.get("model_kwargs", {}),
    )

    for each_task in tqdm(tasks):
        metadata = each_task["metadata"]
        image_description = metadata["generated_description"]

        # extract object descriptions
        obj_description = llm.extract_object_description(image_description)
        # generate what it is
        what_it_is = llm.distill_what_it_is(image_description)
        # extract parts list
        parts_list = llm.extract_parts_list(image_description)
        # extract building steps
        building_steps = llm.extract_building_steps(image_description)

        logging.debug(f"Object Description:\n{obj_description}")
        logging.debug("-" * 50)
        logging.debug(f"What it is:\n{what_it_is}")
        logging.debug("-" * 50)
        logging.debug(f"Parts List:\n{parts_list}")
        logging.debug("-" * 50)
        logging.debug(f"Building Steps:\n{building_steps}")

        if args.save:
            update_metadata = {
                "object_description": obj_description,
                "what_it_is": what_it_is,
                "parts_list": parts_list,
                "building_steps": building_steps,
            }
            save_descriptions(each_task["base_path"], update_metadata)


if __name__ == "__main__":
    _main()



import argparse
import logging
import re
from typing import Any, Dict, Tuple

from cicada.core import model
from cicada.core.basics import PromptBuilder
from cicada.core.utils import colorstring, cprint

logger = logging.getLogger(__name__)


class FeedbackJudge(model.MultiModalModel):
    def __init__(
        self,
        api_key,
        api_base_url,
        model_name,
        org_id,
        prompt_templates,
        **model_kwargs,
    ):
        super().__init__(
            api_key,
            api_base_url,
            model_name,
            org_id,
            **model_kwargs,
        )
        self.prompt_templates = prompt_templates

    def is_design_goal_achieved(
        self, feedback: str, design_goal: str
    ) -> Tuple[bool, float]:
        """
        Determine if the design goal has been achieved and provide a score based on the Hits.

        :param feedback: The feedback to process.
        :param design_goal: The design goal to compare against.
        :return: A tuple containing a boolean indicating if the goal is achieved and a score between 0 and 1.
        """
        feedback_hits = self._extract_hits(feedback)
        cprint(f"Feedback Hits:\n{feedback_hits}", "magenta")
        # Construct the prompt
        prompt = self.prompt_templates["is_design_goal_achieved"][
            "prompt_template"
        ].format(
            design_goal=design_goal,
            feedback_hits=feedback_hits,
        )

        pb = PromptBuilder()
        pb.add_system_message(
            self.prompt_templates["is_design_goal_achieved"]["system_prompt"]
        )
        pb.add_user_message(prompt)

        # Query the LLM
        response = self.query(prompt_builder=pb, stream=self.stream)["content"]
        logger.info(colorstring(f"Feedback Judge determines that:\n{response}", "cyan"))

        # Parse the Markdown-formatted response
        achieved = (
            self._parse_markdown_response(response, key="Achieved").lower() == "yes"
        )
        score = float(self._parse_markdown_response(response, key="Score"))
        return achieved, score

    def _extract_hits(self, feedback: str) -> str:
        """
        Extract the Hits section from the feedback.

        :param feedback: The feedback to parse.
        :return: The Hits section of the feedback, formatted as a string.
        """

        if feedback:
            hits_match = re.search(
                r"##\s*Hits:\s*([\s\S]+?)(?:##|$)", feedback, re.IGNORECASE
            )
            return hits_match.group(1).strip() if hits_match else ""
        return ""

    def _parse_markdown_response(self, response: str, key: str) -> str:
        """
        Parse a Markdown-formatted response to extract the value for a specific key.
        Strips all Markdown formatting (e.g., **, *, _) from the extracted value.

        :param response: The Markdown-formatted response.
        :param key: The key to extract.
        :return: The value associated with the key, stripped of Markdown formatting.
        """

        match = re.search(rf"{key}:\s*(.*)", response, re.IGNORECASE)
        if match:
            value = match.group(1).strip()
            value = re.sub(r"[\\*\\_\\`]", "", value)
            return value
        raise ValueError(f"Key '{key}' not found in response.")


def parse_args() -> Dict[str, Any]:
    """
    Parse command line arguments.

    Returns:
        Dict[str, Any]: Parsed arguments.
    """
    parser = argparse.ArgumentParser(description="Feedback Judge")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    parser.add_argument(
        "--prompts", default="prompts", help="Path to the prompts YAML file or folder"
    )
    parser.add_argument(
        "--design_goal",
        required=True,
        help="Text description of the design goal or path to a JSON file containing the design goal",
    )
    parser.add_argument(
        "--feedback",
        required=True,
        help="Path to the file containing feedback",
    )
    return vars(parser.parse_args())


if __name__ == "__main__":
    from cicada.core.utils import (
        load_config,
        load_prompts,
        parse_design_goal,
        setup_logging,
    )

    args = parse_args()
    setup_logging()

    # Load configuration
    config = load_config(args["config"], "feedback_judge")
    prompt_templates = load_prompts(args["prompts"], "feedback_judge")

    # Initialize FeedbackJudge
    feedback_judge = FeedbackJudge(
        config["api_key"],
        config.get("api_base_url"),
        config.get("model_name", "gpt-4"),
        config.get("org_id"),
        prompt_templates,
        **config.get("model_kwargs", {}),
    )

    # Load design goal
    design_goal = parse_design_goal(args["design_goal"])

    # Load feedback
    with open(args["feedback"], "r") as f:
        feedback = f.read().strip()

    # Evaluate if the design goal is achieved in the feedback
    is_achieved, score = feedback_judge.is_design_goal_achieved(feedback, design_goal)
    print(f"Is design goal achieved? {is_achieved} (Score: {score})")



import logging
import os
from abc import ABC
from typing import Dict, List

import httpx
from cicada.core.utils import colorstring

logger = logging.getLogger(__name__)


class Rerank(ABC):
    def __init__(
        self,
        api_key: str,
        api_base_url: str = "https://api.siliconflow.cn/v1",
        model_name: str = "BAAI/bge-reranker-v2-m3",
        **model_kwargs,
    ):
        """
        Initialize the Rerank class.

        Args:
            api_key (str): API key for authentication.
            api_base_url (str, optional): Base URL for the rerank API. Defaults to "https://api.siliconflow.cn/v1".
            model_name (str, optional): Name of the rerank model. Defaults to "BAAI/bge-reranker-v2-m3".
            **model_kwargs: Additional model-specific parameters.
        """
        self.api_key = api_key
        self.api_base_url = os.path.join(api_base_url, "rerank")
        self.model_name = model_name
        self.model_kwargs = model_kwargs

    def rerank(
        self,
        query: str,
        documents: List[str],
        top_n: int = 4,
        return_documents: bool = False,
    ) -> List[Dict]:
        """
        Rerank a list of documents based on a query.

        Args:
            query (str): The query to rerank documents against.
            documents (List[str]): List of documents to rerank.
            top_n (int, optional): Number of top documents to return. Defaults to 4.
            return_documents (bool, optional): Whether to return the full documents or just scores. Defaults to False.

        Returns:
            List[Dict]: List of reranked documents or scores.
        """
        payload = {
            "model": self.model_name,
            "query": query,
            "documents": documents,
            "top_n": top_n,
            "return_documents": return_documents,
            **self.model_kwargs,
        }
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

        logger.debug(colorstring(f"Payload: {payload}", "blue"))
        logger.debug(colorstring(f"Headers: {headers}", "blue"))
        try:
            response = httpx.post(self.api_base_url, json=payload, headers=headers)
            response.raise_for_status()
            return response.json()["results"]
        except httpx.RequestError as e:
            logger.error(colorstring(f"Failed to rerank documents: {e}", "red"))
            raise


if __name__ == "__main__":
    import argparse

    from cicada.core.utils import colorstring, load_config, setup_logging

    parser = argparse.ArgumentParser(description="Reranking Model")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    args = parser.parse_args()

    setup_logging()

    rerank_config = load_config(args.config, "rerank")

    rerank = Rerank(
        api_key=rerank_config["api_key"],
        api_base_url=rerank_config.get(
            "api_base_url", "https://api.siliconflow.cn/v1/"
        ),
        model_name=rerank_config.get("model_name", "BAAI/bge-reranker-v2-m3"),
        **rerank_config.get("model_kwargs", {}),
    )

    query = "Apple"
    documents = ["苹果", "香蕉", "水果", "蔬菜"]
    reranked_results = rerank.rerank(query, documents, top_n=4, return_documents=False)
    logger.info(colorstring(f"Reranked results: {reranked_results}", "white"))



import os
import sys
import uuid

import yaml
from pgvector.sqlalchemy import Vector
from sqlalchemy import (
    TEXT,
    TIMESTAMP,
    Column,
    Enum,
    ForeignKey,
    MetaData,
    String,
    Table,
    create_engine,
)
from sqlalchemy.dialects.postgresql import UUID as PGUUID
from sqlalchemy.schema import CreateSchema

EMBEDDING_SIZE = 1024

_current_dir = os.path.dirname(os.path.abspath(__file__))

# Load configuration from config.yaml
with open(os.path.join(_current_dir, "config.yaml"), "r") as file:
    config = yaml.safe_load(file)

db_config = config["db_config"]

# Extract database configuration
DATABASE_URL = "postgresql://{user}:{password}@{host}:{port}/{database}".format(
    user=db_config["user"],
    password=db_config["password"],
    host=db_config["host"],
    port=db_config["port"],
    database=db_config["database"],
)

# Create an engine instance
engine = create_engine(DATABASE_URL)

# Create the schema if it doesn't exist
with engine.connect() as connection:
    if not engine.dialect.has_schema(connection, db_config["schema"]):
        connection.execute(CreateSchema(db_config["schema"]))
        connection.commit()

# Define the metadata
metadata = MetaData(schema=db_config["schema"])

# Define the tables
cad_objects = Table(
    "cad_objects",
    metadata,
    Column("id", PGUUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
    Column("name", String(255)),
    Column("created_at", TIMESTAMP),
    Column("updated_at", TIMESTAMP),
    Column(
        "parent_object_id",
        PGUUID(as_uuid=True),
        ForeignKey("cad_objects.id"),
        nullable=True,
    ),
)

files = Table(
    "files",
    metadata,
    Column("id", PGUUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
    Column("cad_object_id", PGUUID(as_uuid=True), ForeignKey("cad_objects.id")),
    Column("file_type", Enum("STEP", "OBJ", "PLY", name="file_type_enum")),
    Column("file_path", String(255)),
    Column("created_at", TIMESTAMP),
    Column("updated_at", TIMESTAMP),
)

descriptions = Table(
    "descriptions",
    metadata,
    Column("id", PGUUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
    Column("cad_object_id", PGUUID(as_uuid=True), ForeignKey("cad_objects.id")),
    Column(
        "description_type",
        Enum("LLM_GENERATED", "DISTILLED", name="description_type_enum"),
    ),
    Column("text", TEXT),
    Column("description_embedding", Vector(EMBEDDING_SIZE)),
    Column("created_at", TIMESTAMP),
    Column("updated_at", TIMESTAMP),
)

cadquery_codes = Table(
    "cadquery_codes",
    metadata,
    Column("id", PGUUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
    Column("cad_object_id", PGUUID(as_uuid=True), ForeignKey("cad_objects.id")),
    Column("code", TEXT),
    Column("code_embedding", Vector(EMBEDDING_SIZE)),
    Column("created_at", TIMESTAMP),
    Column("updated_at", TIMESTAMP),
)

multiview_snapshots = Table(
    "multiview_snapshots",
    metadata,
    Column("id", PGUUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
    Column("cad_object_id", PGUUID(as_uuid=True), ForeignKey("cad_objects.id")),
    Column("image_path", String(255)),
    Column("image_embedding", Vector(EMBEDDING_SIZE)),
    Column("created_at", TIMESTAMP),
    Column("updated_at", TIMESTAMP),
)

object_parts = Table(
    "object_parts",
    metadata,
    Column("id", PGUUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
    Column("parent_object_id", PGUUID(as_uuid=True), ForeignKey("cad_objects.id")),
    Column("part_object_id", PGUUID(as_uuid=True), ForeignKey("cad_objects.id")),
    Column("created_at", TIMESTAMP),
    Column("updated_at", TIMESTAMP),
)

# Create the tables
metadata.create_all(engine)



from typing import List, Union

from cicada.core.embed import Embed
from cicada.retrieval.basics import Document, Embeddings


class SiliconFlowEmbeddings(Embeddings):
    """Embedding class for SiliconFlow BGE-M3 API."""

    def __init__(
        self,
        api_key: str,
        api_base_url: str,
        model_name: str,
        org_id: str = None,
        **model_kwargs,
    ):
        """
        Initialize the SiliconFlow BGE-M3 embedding model.

        Args:
            api_key (str): The API key for SiliconFlow.
            api_base_url (str): The base URL for the API.
            model_name (str): The name of the model to use.
            org_id (str, optional): The organization ID. Defaults to None.
            **model_kwargs: Additional keyword arguments for the model.
        """
        self.api_key = api_key
        self.api_base_url = api_base_url
        self.model_name = model_name
        self.org_id = org_id
        self.model_kwargs = model_kwargs

        # Initialize the Embed class from embed.py
        self.embed_client = Embed(
            api_key=self.api_key,
            api_base_url=self.api_base_url,
            model_name=self.model_name,
            org_id=self.org_id,
            **self.model_kwargs,
        )

    def embed_query(self, text: str) -> List[float]:
        """
        Generate an embedding for a single query text.

        Args:
            text (str): The query text to embed.

        Returns:
            List[float]: The embedding of the query text.
        """
        return self.embed_documents(text)[0]

    def embed_documents(
        self, texts: Union[List[str], str, Document, List[Document]]
    ) -> List[List[float]]:
        """
        Generate embeddings for a list of texts or documents.

        Args:
            texts (Union[List[str], str, Document, List[Document]]): The input texts or documents to embed.

        Returns:
            List[List[float]]: A list of embeddings for the input texts or documents.

        Raises:
            ValueError: If the input list contains mixed types of strings and Documents.
            TypeError: If the input is not a string, Document, list of strings, or list of Documents.
        """
        # Normalize input to a list of strings
        if isinstance(texts, str):
            texts = [texts]
        elif isinstance(texts, Document):
            texts = [texts.page_content]
        elif isinstance(texts, list):
            if all(isinstance(text, str) for text in texts):
                pass  # Already a list of strings
            elif all(isinstance(text, Document) for text in texts):
                texts = [text.page_content for text in texts]
            else:
                raise ValueError(
                    "Input list must contain only strings or only Document instances."
                )
        else:
            raise TypeError(
                "Input must be a string, Document, list of strings, or list of Documents."
            )

        # Generate embeddings using the Embed client
        return self.embed_client.embed(texts)


if __name__ == "__main__":
    import argparse

    from cicada.core.utils import load_config

    parser = argparse.ArgumentParser(description="Feedback Judge")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    args = parser.parse_args()

    embed_config = load_config(args.config, "embed")

    embedding_model = SiliconFlowEmbeddings(
        embed_config["api_key"],
        embed_config.get("api_base_url"),
        embed_config.get("model_name", "text-embedding-3-small"),
        embed_config.get("org_id"),
        **embed_config.get("model_kwargs", {}),
    )

    # Generate embeddings for a list of texts
    texts = ["This is a test document.", "Another test document."]
    embeddings = embedding_model.embed_documents(texts)
    print(embeddings)

    # Generate an embedding for a single query
    query_embedding = embedding_model.embed_query("This is a query.")
    print(query_embedding)



import ast
import importlib.util
import logging
import os
import shutil
import subprocess
import tempfile
from subprocess import run, TimeoutExpired

from cicada.core.utils import colorstring

logger = logging.getLogger(__name__)


class CodeExecutor:
    def execute_code(
        self,
        code: str,
        timeout: int = 10,
        test_run: bool = False,
    ) -> tuple[bool, dict]:
        """
        Executes the provided Python code in a temporary directory and returns the results.

        Args:
            code (str): The Python code to execute.
            timeout (int, optional): The maximum time (in seconds) to allow for the execution. Defaults to 10.

        Returns:
            tuple[bool, dict]: A tuple where the first element is a boolean indicating success (True) or failure (False), and the second element is a dictionary containing the results. If successful, the dictionary will contain the standard output and any generated files. If unsuccessful, it will contain an error message.
        """
        temp_dir = tempfile.mkdtemp()
        script_path = os.path.join(temp_dir, "script.py")
        with open(script_path, "w") as f:
            f.write(code)

        try:
            completed_process = run(
                ["python", script_path],
                cwd=temp_dir,
                capture_output=True,
                text=True,
                timeout=timeout,
            )
            if completed_process.returncode != 0:
                error_message = completed_process.stderr
                return False, {"error": error_message}
            else:
                if test_run:
                    logger.info("Test run successful.")
                    return True, {"output": completed_process.stdout}
                else:
                    # Collect all files generated during execution.
                    output_files = {}
                    for root, _, files in os.walk(temp_dir):
                        for filename in files:
                            file_path = os.path.join(root, filename)
                            rel_path = os.path.relpath(file_path, temp_dir)
                            with open(file_path, "rb") as f:
                                content = f.read()
                                output_files[rel_path] = content
                    logger.info("Execution successful. Collecting output files.")
                    return True, {
                        "output": completed_process.stdout,
                        "files": output_files,
                    }
        except TimeoutExpired:
            return False, {"error": "Execution timed out."}
        finally:
            shutil.rmtree(temp_dir)

    def execute_and_save(
        self,
        code: str,
        output_path: str,
        certain_file_types: list[str] | None = None,
        timeout: int = 10,
    ) -> tuple[bool, str]:
        """
        Executes the provided code and saves the output artifacts to the specified directory.

        Args:
            code (str): The Python code to execute.
            output_path (str): The directory where the output artifacts should be saved.
            certain_file_types (list, optional): A list of file extensions to save. If None, all files are saved.
            timeout (int, optional): The maximum execution time in seconds. Defaults to 10 seconds.

        Returns:
            tuple: A tuple containing a boolean indicating success and a message string.
        """
        success, result = self.execute_code(code, timeout=timeout)
        if not success:
            return False, result["error"]

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        files = result.get("files", {})
        logging.debug(f"Output files: {files}")
        for file_path, content in files.items():
            if certain_file_types is not None and not file_path.endswith(
                tuple(certain_file_types)
            ):
                continue
            file_output_path = os.path.join(output_path, file_path)
            os.makedirs(os.path.dirname(file_output_path), exist_ok=True)
            with open(file_output_path, "wb") as f:
                f.write(content)

        stdout = result.get("output", "")
        output_log_path = os.path.join(output_path, "output.log")
        with open(output_log_path, "w") as f:
            f.write(stdout)

        return True, "Artifacts saved successfully to {}".format(output_path)

    def check_syntax(self, code: str) -> tuple[bool, str | None]:
        try:
            compile(code, "<string>", "exec")
            return True, None
        except SyntaxError as e:
            return False, str(e)

    def check_grammar(self, code: str) -> tuple[bool, str | None]:
        try:
            ast.parse(code)
            return True, None
        except SyntaxError as e:
            return False, str(e)

    def check_imports(self, code: str) -> tuple[bool, str | None]:
        missing_modules = []
        lines = code.splitlines()
        for line in lines:
            if line.startswith("import ") or line.startswith("from "):
                module = line.split()[1].split(".")[0]  # Extract the module name
                if importlib.util.find_spec(module) is None:
                    missing_modules.append(module)

        if missing_modules:
            return False, f"Missing Modules: {', '.join(missing_modules)}"
        else:
            return True, None

    def validate_code(self, code: str) -> tuple[bool, str | None]:
        syntax_result, syntax_error = self.check_syntax(code)
        if not syntax_result:
            return False, f"Syntax Error: {syntax_error}"

        grammar_result, grammar_error = self.check_grammar(code)
        if not grammar_result:
            return False, f"Grammar Error: {grammar_error}"

        imports_result, imports_error = self.check_imports(code)
        if not imports_result:
            return False, f"Import Error: {imports_error}"

        return True, None


if __name__ == "__main__":
    from cicada.core.utils import colorstring, setup_logging

    setup_logging()

    code_executor = CodeExecutor()

    # code_file = (
    #     "/home/pding/projects/codecad/codecad-rag/code-generate/reproduce-mech-part.py"
    # )
    # with open(code_file, "r") as f:
    #     code = f.read()

    # result = code_executor.execute_code(code)
    # print(result)

    # Test cases
    test_cases = [
        {
            "name": "Valid Python Code",
            "code": """
def add(a, b):
    return a + b

result = add(1, 2)
print(result)
""",
        },
        {
            "name": "Syntax Error: Missing Colon",
            "code": """
def add(a, b)
    return a + b

result = add(1, 2)
print(result)
""",
        },
        {
            "name": "Syntax Error: Indentation Error",
            "code": """
def add(a, b):
return a + b

result = add(1, 2)
print(result)
""",
        },
        {
            "name": "Syntax Error: Invalid Syntax",
            "code": """
if True
    print("Hello, World!")
""",
        },
    ]

    for case in test_cases:
        print(f"Testing: {case['name']}")
        is_valid, error_message = code_executor.validate_code(case["code"])
        if is_valid:
            print("Syntax is valid.")
        else:
            print(f"Syntax error detected: {error_message}")
        print("-" * 40)

    with open("/tmp/cicada/code_examples/patched_code.py") as f:
        patched_code = f.read()

    code_executor.execute_and_save(patched_code, "/tmp/cicada/stl_examples")



import logging
import uuid
from typing import Dict, List, Optional, Tuple

import sqlalchemy
from pgvector.sqlalchemy import Vector
from sqlalchemy import Column, create_engine, ForeignKey, JSON, String, UUID
from sqlalchemy.orm import declarative_base, relationship, Session, sessionmaker

from cicada.core.utils import colorstring, cprint, load_config, setup_logging
from cicada.retrieval.basics import Document, Embeddings, VectorStore

logger = logging.getLogger(__name__)

Base = declarative_base()


class CollectionStore(Base):
    """Represents a collection in the database.

    Attributes:
        uuid (UUID): Primary key for the collection.
        name (str): Name of the collection.
        cmetadata (dict): Metadata associated with the collection.
        embeddings (List[EmbeddingStore]): List of embeddings associated with the collection.
    """

    __tablename__ = "vectorsearch_pg_collection"
    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(String)
    cmetadata = Column(JSON)
    embeddings = relationship(
        "EmbeddingStore", back_populates="collection", passive_deletes=True
    )


class EmbeddingStore(Base):
    """Represents an embedding in the database.

    Attributes:
        uuid (UUID): Primary key for the embedding.
        collection_id (UUID): Foreign key referencing the collection.
        collection (CollectionStore): Collection associated with the embedding.
        embedding (Vector): The embedding vector.
        document (str): The document associated with the embedding.
        cmetadata (dict): Metadata associated with the embedding.
        custom_id (str): Custom ID for the embedding.
    """

    __tablename__ = "vectorsearch_pg_embedding"
    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    collection_id = Column(
        UUID(as_uuid=True),
        ForeignKey("vectorsearch_pg_collection.uuid", ondelete="CASCADE"),
    )
    collection = relationship(CollectionStore, back_populates="embeddings")
    embedding = Column(Vector)
    document = Column(String, nullable=True)
    cmetadata = Column(JSON, nullable=True)
    custom_id = Column(String, nullable=True)


class PGVector(VectorStore):
    """A vector store implementation using PostgreSQL and pgvector.

    Attributes:
        _engine (sqlalchemy.engine.Engine): SQLAlchemy engine for database connection.
        _Session (sqlalchemy.orm.sessionmaker): SQLAlchemy session maker.
        _embedding (Embeddings): Embedding model used for generating embeddings.
        _collection_name (str): Name of the collection in the database.
        _collection (CollectionStore): The collection associated with this vector store.
    """

    def __init__(
        self,
        connection_string: str,
        embedding: Embeddings,
        collection_name: str = "vectorsearch",
        pool_size: int = 5,
        **kwargs,
    ):
        """Initialize the PGVector store.

        Args:
            connection_string (str): Database connection string.
            embedding (Embeddings): Embedding model to use.
            collection_name (str, optional): Name of the collection. Defaults to "vectorsearch".
            pool_size (int, optional): Connection pool size. Defaults to 5.
        """
        self._engine = create_engine(
            connection_string, pool_size=pool_size, max_overflow=10
        )
        self._Session = sessionmaker(bind=self._engine)
        self._embedding = embedding
        self._collection_name = collection_name
        self.create_tables_if_not_exists()
        self._collection = self._get_or_create_collection()
        # Validate embedding dimensions
        self._validate_embedding_dimensionality()

    def _validate_embedding_dimensionality(self):
        """Check if embedding dimensions match the database schema."""
        dummy_embedding = self._embedding.embed_query("test")
        column_type = EmbeddingStore.embedding.type
        if hasattr(column_type, "dimensions") and column_type.dimensions is not None:
            if len(dummy_embedding) != column_type.dimensions:
                raise ValueError(
                    f"Embedding dimension ({len(dummy_embedding)}) "
                    f"does not match database schema ({column_type.dimensions})."
                )

    def _get_or_create_collection(self) -> CollectionStore:
        """Get existing collection or create a new one.
        Returns:
            CollectionStore: The collection object.
        """
        session = self._Session()
        try:
            collection = (
                session.query(CollectionStore)
                .filter_by(name=self._collection_name)
                .first()
            )
            if not collection:
                collection = CollectionStore(name=self._collection_name, cmetadata={})
                session.add(collection)
                session.commit()
            return collection
        except Exception as e:
            session.rollback()
            raise e
        finally:
            session.close()

    def set_collection_metadata(self, key: str, value: str):
        """Set metadata for the collection.

        Args:
            key (str): Metadata key.
            value (str): Metadata value.
        """
        with self._Session() as session:
            # Using session.merge to re-attach the detached instance
            collection = session.merge(self._collection)
            if not collection.cmetadata:
                collection.cmetadata = {}
            collection.cmetadata[key] = value
            session.commit()

    def get_collection_metadata(self, key: str) -> Optional[str]:
        """Get metadata from the collection.

        Args:
            key (str): Metadata key.

        Returns:
            Optional[str]: The metadata value if it exists, otherwise None.
        """
        with self._Session() as session:
            # Re-fetch the collection within the new session context
            # ===== NOT Correct ========
            # collection = session.get(CollectionStore, self._collection.uuid)
            #
            # ===== Correct ========
            # collection = (
            #     session.query(CollectionStore)
            #     .filter_by(name=self._collection_name)
            #     .first()
            # )
            #
            # ===== Why? ========
            # The original code was incorrect because it directly accessed self._collection.uuid,
            # which could be detached from the session.
            collection = (
                session.query(CollectionStore)
                .filter_by(name=self._collection_name)
                .first()
            )
            return collection.cmetadata.get(key) if collection.cmetadata else None

    def get_all_collection_metadata(self) -> Dict:
        """Get all metadata from the collection.

        Returns:
            Dict: A dictionary containing all metadata.
        """
        with self._Session() as session:
            # Re-fetch the collection within the new session context (see above for explanation)
            collection = (
                session.query(CollectionStore)
                .filter_by(name=self._collection_name)
                .first()
            )
            return collection.cmetadata or {}

    def create_tables_if_not_exists(self) -> None:
        """Create tables in the database if they don't exist."""
        try:
            Base.metadata.create_all(self._engine)
            logger.info(colorstring("Tables created or verified in PostgreSQL", "blue"))
        except Exception as e:
            logger.error(colorstring(f"Failed to create tables: {e}", "red"))
            raise e

    def create_collection(self) -> None:
        """Create a collection in the database."""
        session = self._Session()
        try:
            collection = CollectionStore(name=self._collection_name)
            session.add(collection)
            session.commit()
            logger.info(
                colorstring(f"Collection '{self._collection_name}' created", "blue")
            )
        except Exception as e:
            session.rollback()
            logger.error(colorstring(f"Failed to create collection: {e}", "red"))
            raise e
        finally:
            session.close()

    def delete_by_ids(self, ids: List[str]):
        """Delete documents by their IDs."""
        session = self._Session()
        try:
            session.query(EmbeddingStore).filter(
                EmbeddingStore.custom_id.in_(ids)
            ).delete()
            session.commit()
            logger.info(colorstring(f"Deleted {len(ids)} documents", "blue"))
        except Exception as e:
            session.rollback()
            logger.error(colorstring(f"Failed to delete documents: {e}", "red"))
            raise e
        finally:
            session.close()

    def add_texts(
        self, texts: List[str], metadatas: Optional[List[Dict]] = None
    ) -> List[str]:
        """Add texts to the vector store.

        Args:
            texts (List[str]): The texts to add.
            metadatas (Optional[List[Dict]], optional): Metadata for each text. Defaults to None.

        Returns:
            List[str]: The IDs of the added texts.
        """
        session = self._Session()
        try:
            embeddings = self._embedding.embed_documents(texts)
            metadatas = metadatas or [{} for _ in texts]
            ids = [str(uuid.uuid4()) for _ in texts]
            collection = (
                session.query(CollectionStore)
                .filter_by(name=self._collection_name)
                .first()
            )
            documents = [
                EmbeddingStore(
                    embedding=embedding,
                    document=text,
                    cmetadata=metadata,
                    custom_id=id,
                    collection_id=collection.uuid,
                )
                for text, metadata, embedding, id in zip(
                    texts, metadatas, embeddings, ids
                )
            ]
            session.bulk_save_objects(documents)
            session.commit()
            logger.info(
                colorstring(f"Added {len(texts)} texts to the vector store", "blue")
            )
            return ids
        except Exception as e:
            session.rollback()
            logger.error(colorstring(f"Failed to add texts: {e}", "red"))
            raise e
        finally:
            session.close()

    def add_texts_with_embeddings(
        self,
        texts: List[str],
        embeddings: List[List[float]],
        metadatas: Optional[List[Dict]] = None,
    ) -> List[str]:
        """Add texts with precomputed embeddings to the vector store.

        Args:
            texts (List[str]): The texts to add.
            embeddings (List[List[float]]): Precomputed embeddings.
            metadatas (Optional[List[Dict]], optional): Metadata for each text. Defaults to None.

        Returns:
            List[str]: The IDs of the added texts.
        """
        if len(texts) != len(embeddings):
            raise ValueError("Texts and embeddings must have the same length.")

        session = self._Session()
        try:
            metadatas = metadatas or [{} for _ in texts]
            ids = [str(uuid.uuid4()) for _ in texts]
            collection = (
                session.query(CollectionStore)
                .filter_by(name=self._collection_name)
                .first()
            )
            documents = [
                EmbeddingStore(
                    embedding=embedding,
                    document=text,
                    cmetadata=metadata,
                    custom_id=id,
                    collection_id=collection.uuid,
                )
                for text, metadata, embedding, id in zip(
                    texts, metadatas, embeddings, ids
                )
            ]
            session.bulk_save_objects(documents)
            session.commit()
            logger.info(
                colorstring(f"Added {len(texts)} texts with custom embeddings", "blue")
            )
            return ids
        except Exception as e:
            session.rollback()
            logger.error(colorstring(f"Failed to add texts: {e}", "red"))
            raise e
        finally:
            session.close()

    def similarity_search(
        self, query: str, k: int = 4
    ) -> Tuple[List[Document], List[float]]:
        """Perform a similarity search for a query.

        Args:
            query (str): The query to search for.
            k (int, optional): The number of results to return. Defaults to 4.

        Returns:
            Tuple[List[Document], List[float]]: A tuple containing the list of documents that match the query and their corresponding similarity scores.
        """
        try:
            embedding = self._embedding.embed_query(query)
            logger.info(
                colorstring(f"Performing similarity search for query: {query}", "cyan")
            )
            return self.similarity_search_by_vector(embedding, k)
        except Exception as e:
            logger.error(
                colorstring(f"Failed to perform similarity search: {e}", "red")
            )
            raise e

    def similarity_search_by_vector(
        self, embedding: List[float], k: int = 4, distance_metric: str = "cosine"
    ) -> Tuple[List[Document], List[float]]:
        """Perform a similarity search by vector with configurable distance metrics.

        Args:
            embedding (List[float]): The embedding vector to search with.
            k (int, optional): The number of results to return. Defaults to 4.
            distance_metric (str, optional): The distance metric to use. Defaults to "l2". Options are "l2" and "cosine". see https://github.com/pgvector/pgvector?tab=readme-ov-file#querying for more details.

        Returns:
            Tuple[List[Document], List[float]]: Documents and similarity scores.
        """
        # Validate distance metric
        if distance_metric not in ["l2", "cosine"]:
            raise ValueError(
                f"Unsupported distance metric: {distance_metric}. Use 'l2' or 'cosine'."
            )

        # Normalization check (same for both implementations)
        if distance_metric == "cosine":
            l2_norm = sum(x**2 for x in embedding) ** 0.5
            if not (0.99 < l2_norm < 1.01):
                distance_metric = "l2"
                logger.warning(
                    colorstring("Non-unit vector - using L2 distance", "yellow")
                )

        session = self._Session()
        try:
            collection = (
                session.query(CollectionStore)
                .filter_by(name=self._collection_name)
                .first()
            )
            match distance_metric:
                case "cosine":
                    distance_expr = EmbeddingStore.embedding.cosine_distance(embedding)
                case "l2":
                    distance_expr = EmbeddingStore.embedding.l2_distance(embedding)
                case _:
                    raise ValueError("Invalid distance metric. Use 'l2', or 'cosine'.")

            results = (
                session.query(EmbeddingStore, distance_expr.label("distance"))
                .filter(EmbeddingStore.collection_id == collection.uuid)
                .order_by(distance_expr)
                .limit(k)
                .all()
            )
            docs = [
                Document(
                    page_content=result.EmbeddingStore.document,
                    metadata=result.EmbeddingStore.cmetadata,
                )
                for result in results
            ]
            scores = [result.distance for result in results]
            logger.info(
                colorstring(
                    f"Found {len(docs)} results using {distance_metric} metric",
                    "cyan",
                )
            )
            return docs, scores
        except Exception as e:
            logger.error(
                colorstring(f"Similarity search failed ({distance_metric}): {e}", "red")
            )
            raise e
        finally:
            session.close()


if __name__ == "__main__":
    import argparse

    from cicada.retrieval.siliconflow_embeddings import SiliconFlowEmbeddings
    from cicada.retrieval.siliconflow_rerank import SiliconFlowRerank

    setup_logging()

    parser = argparse.ArgumentParser(description="Feedback Judge")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    args = parser.parse_args()

    embed_config = load_config(args.config, "embed")

    embedding_model = SiliconFlowEmbeddings(
        embed_config["api_key"],
        embed_config.get("api_base_url"),
        embed_config.get("model_name", "text-embedding-3-small"),
        embed_config.get("org_id"),
        **embed_config.get("model_kwargs", {}),
    )

    rerank_config = load_config(args.config, "rerank")

    rerank_model = SiliconFlowRerank(
        api_key=rerank_config["api_key"],
        api_base_url=rerank_config.get(
            "api_base_url", "https://api.siliconflow.cn/v1/"
        ),
        model_name=rerank_config.get("model_name", "BAAI/bge-reranker-v2-m3"),
        **rerank_config.get("model_kwargs", {}),
    )

    pgvector_store_config = load_config(args.config, "pgvector_store")
    connection_string = pgvector_store_config["connection_string"]
    collection_name = pgvector_store_config["collection_name"]
    pg_vector = PGVector(
        connection_string=connection_string,
        embedding=embedding_model,
        collection_name=collection_name,
        pool_size=5,
    )

    # ============ metadata operations ============
    # Test metadata functionality
    cprint("\nTesting metadata operations...", "cyan")

    # Set metadata
    pg_vector.set_collection_metadata("version", "1.0")
    pg_vector.set_collection_metadata("status", "active")

    # Get single metadata
    version = pg_vector.get_collection_metadata("version")
    cprint(f"Retrieved version: {version}", "green")

    # Get all metadata
    all_metadata = pg_vector.get_all_collection_metadata()
    cprint(f"All metadata: {all_metadata}", "blue")

    # Test non-existent key
    missing = pg_vector.get_collection_metadata("nonexistent")
    cprint(f"Non-existent key returns: {missing}", "yellow")

    # ============ text operations ============
    # Add texts
    texts = [
        "apple",  # English
        "PEAR",  # English (uppercase)
        "naranja",  # Spanish
        "葡萄",  # Chinese
        "The quick brown fox jumps over the lazy dog.",  # English sentence
        "La rápida zorra marrón salta sobre el perro perezoso.",  # Spanish sentence
        "敏捷的棕色狐狸跳过了懒狗。",  # Chinese sentence
        "12345",  # Numbers
        "Café au lait",  # French with special character
        "🍎🍐🍇",  # Emojis
        "manzana",  # Spanish for apple
        "pomme",  # French for apple
        "苹果",  # Chinese for apple
        "grape",  # English for grape
        "uva",  # Spanish for grape
        "fox",  # English for fox
        "zorro",  # Spanish for fox
    ]
    metadatas = [{"source": f"test{i+1}"} for i in range(len(texts))]
    ids = pg_vector.add_texts(texts, metadatas)
    cprint(f"Added texts with IDs: {ids}", "blue")

    # Perform similarity search
    queries = [
        "Vínber",  # Icelandic for grape
        "manzana",  # Spanish for apple
        "狐狸",  # Chinese for fox
        "lazy",  # English word
        "rápida",  # Spanish word
        "🍇",  # Grape emoji
        "Café",  # French word with special character
        "123",  # Partial number
    ]

    for query in queries:
        cprint(f"\nQuery: {query}", "blue")
        results, scores = pg_vector.similarity_search(query, k=10)
        cprint(f"Similarity search results: {list(zip(results, scores))}", "yellow")

        # Rerank the results
        reranked_results = rerank_model.rerank(
            query,
            results,
            top_n=5,
            return_documents=True,
        )
        cprint(f"Reranked results: {reranked_results}", "cyan")

    # Clean up (optional)
    with pg_vector._Session() as session:
        collection = (
            session.query(CollectionStore).filter_by(name=collection_name).first()
        )
        if collection:
            session.delete(collection)
            session.commit()
            cprint(f"Removed test collection: {collection_name}", "yellow")



from typing import List, Dict, Optional
import logging

logger = logging.getLogger(__name__)


class Document:
    """A simple class to hold text and metadata."""

    def __init__(self, page_content: str, metadata: Optional[Dict] = None):
        """
        Initialize a Document instance.
        Args:
            page_content (str): The text content of the document.
            metadata (Optional[Dict]): A dictionary of metadata associated with the document. Defaults to None.
        """
        self.page_content = page_content
        self.metadata = metadata or {}

    def __str__(self):
        """
        Return a string representation of the Document instance.
        Returns:
            str: A string representation of the Document.
        """
        metadata_str = ", ".join(f"{k}: {v}" for k, v in self.metadata.items())
        return (
            f"Document(page_content='{self.page_content}', metadata={{{metadata_str}}})"
        )

    def pretty_print(self, indent: int = 0):
        """
        Return a formatted string representation of the Document instance with optional indentation.
        Args:
            indent (int): The number of spaces to indent the output. Defaults to 0.
        Returns:
            str: A formatted string representation of the Document.
        """
        indent_str = " " * indent
        metadata_str = (
            ",\n"
            + indent_str
            + "    ".join(f"{k}: {v}" for k, v in self.metadata.items())
        )
        return f"{indent_str}Document(\n{indent_str}    page_content='{self.page_content}',\n{indent_str}    metadata={{{metadata_str}\n{indent_str}}})"

    def __repr__(self):
        """
        Return a detailed string representation of the Document instance.

        Returns:
            str: A detailed string representation of the Document.
        """
        return self.__str__()


from typing import List, Union, Dict, Optional


class Embeddings:
    """Base class for embedding models."""

    def embed_query(self, text: str) -> List[float]:
        """
        Generate an embedding for a single query text.

        Args:
            text (str): The text to embed.

        Returns:
            List[float]: The embedding vector.
        """
        raise NotImplementedError

    def embed_documents(
        self, texts: Union[List[str], str, Document, List[Document]]
    ) -> List[List[float]]:
        """
        Generate embeddings for a list of texts, a single text, a Document, or a list of Documents.

        Args:
            texts: Can be a list of strings, a single string, a Document, or a list of Documents.

        Returns:
            List[List[float]]: A list of embeddings, where each embedding is a list of floats.
        """
        raise NotImplementedError


class VectorStore:
    """Base class for vector stores."""

    def add_texts(
        self, texts: List[str], metadatas: Optional[List[Dict]] = None
    ) -> List[str]:
        """
        Add texts to the vector store with optional metadata.

        Args:
            texts (List[str]): A list of texts to add.
            metadatas (Optional[List[Dict]]): A list of metadata dictionaries corresponding to the texts. Defaults to None.

        Returns:
            List[str]: A list of IDs or keys associated with the added texts.
        """
        raise NotImplementedError

    def similarity_search(self, query: str, k: int = 4) -> List[Document]:
        """
        Perform a similarity search for the given query string.

        Args:
            query (str): The query string to search for.
            k (int): The number of results to return. Defaults to 4.

        Returns:
            List[Document]: A list of Document instances that are most similar to the query.
        """
        raise NotImplementedError

    def similarity_search_by_vector(
        self, embedding: List[float], k: int = 4
    ) -> List[Document]:
        """
        Perform a similarity search using a precomputed embedding vector.

        Args:
            embedding (List[float]): The embedding vector to search with.
            k (int): The number of results to return. Defaults to 4.

        Returns:
            List[Document]: A list of Document instances that are most similar to the embedding.
        """
        raise NotImplementedError



from typing import Dict, List, Union

from cicada.core.rerank import Rerank
from cicada.retrieval.basics import Document


class SiliconFlowRerank(Rerank):
    """Rerank class for SiliconFlow BGE-Reranker API."""

    def __init__(
        self,
        api_key: str,
        api_base_url: str = "https://api.siliconflow.cn/v1",
        model_name: str = "BAAI/bge-reranker-v2-m3",
        **model_kwargs,
    ):
        """
        Initialize the SiliconFlow BGE-Reranker model.

        Args:
            api_key (str): The API key for SiliconFlow.
            api_base_url (str, optional): The base URL for the API. Defaults to "https://api.siliconflow.cn/v1".
            model_name (str, optional): The name of the model to use. Defaults to "BAAI/bge-reranker-v2-m3".
            **model_kwargs: Additional keyword arguments for the model.
        """
        super().__init__(
            api_key=api_key,
            api_base_url=api_base_url,
            model_name=model_name,
            **model_kwargs,
        )

    def rerank(
        self,
        query: str,
        texts: Union[List[str], str, Document, List[Document]],
        top_n: int = 4,
        return_documents: bool = False,
    ) -> List[Dict]:
        """
        Rerank a list of texts or documents based on a query.

        Args:
            query (str): The query to rerank texts or documents against.
            texts (Union[List[str], str, Document, List[Document]]): The input texts or documents to rerank.
            top_n (int, optional): Number of top documents to return. Defaults to 4.
            return_documents (bool, optional): Whether to return the full documents or just scores. Defaults to False.

        Returns:
            List[Dict]: List of reranked documents or scores.

        Raises:
            ValueError: If the input list contains mixed types of strings and Documents.
            TypeError: If the input is not a string, Document, list of strings, or list of Documents.
        """
        # Normalize input to a list of strings
        if isinstance(texts, str):
            texts = [texts]
        elif isinstance(texts, Document):
            texts = [texts.page_content]
        elif isinstance(texts, list):
            if all(isinstance(text, str) for text in texts):
                pass  # Already a list of strings
            elif all(isinstance(text, Document) for text in texts):
                texts = [text.page_content for text in texts]
            else:
                raise ValueError(
                    "Input list must contain only strings or only Document instances."
                )
        else:
            raise TypeError(
                "Input must be a string, Document, list of strings, or list of Documents."
            )

        # Call the parent class's rerank method with the normalized list of texts
        return super().rerank(query, texts, top_n, return_documents)


if __name__ == "__main__":

    import argparse

    from cicada.core.utils import load_config, setup_logging

    setup_logging()

    parser = argparse.ArgumentParser(description="Feedback Judge")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    args = parser.parse_args()

    rerank_config = load_config(args.config, "rerank")

    rerank_model = SiliconFlowRerank(
        api_key=rerank_config["api_key"],
        api_base_url=rerank_config.get(
            "api_base_url", "https://api.siliconflow.cn/v1/"
        ),
        model_name=rerank_config.get("model_name", "BAAI/bge-reranker-v2-m3"),
        **rerank_config.get("model_kwargs", {}),
    )

    # Rerank a list of documents based on a query
    query = "Apple的同义词"
    documents = ["苹果", "banana", "水果", "vegi", "manzana"]
    reranked_results = rerank_model.rerank(
        query, documents, top_n=4, return_documents=True
    )
    print(reranked_results)



import json
import os
from typing import List

from cicada.core.utils import load_config


def get_images_from_folder(folder_path: str) -> List[str]:
    """Get all image files from the specified folder."""
    valid_extensions = [".jpg", ".jpeg", ".png", ".bmp", ".gif"]
    images = [
        os.path.join(folder_path, f)
        for f in os.listdir(folder_path)
        if os.path.splitext(f)[1].lower() in valid_extensions
    ]
    return images


def save_descriptions(base_dirs: str | List[str], descriptions: dict | List[dict]):

    if not isinstance(base_dirs, list):
        base_dirs = [base_dirs]
    if not isinstance(descriptions, list):
        descriptions = [descriptions]

    assert len(base_dirs) == len(
        descriptions
    ), "Number of base paths and descriptions do not match"

    for directory, desc in zip(base_dirs, descriptions):
        metadata = {}
        metadata_path = os.path.join(directory, "metadata.json")
        if os.path.exists(metadata_path):
            metadata = load_config(metadata_path)

        # extend metadata.json with the new descriptions
        metadata.update(desc)

        with open(metadata_path, "w") as file:
            json.dump(metadata, file, indent=4)


def load_object_metadata(task_path: str) -> List[dict]:
    if os.path.isdir(task_path):
        directory = task_path
        image_metadata = {
            "object_id": os.path.basename(directory),
            "object_description": "",
            "base_path": directory,
            "images": [],
        }
        for file in os.listdir(directory):
            if file.lower().endswith((".png", ".jpg", ".jpeg")):
                pre_description = (
                    file.split(".")[0].replace("-", " ").replace("_", " ").strip()
                )
                image_metadata["images"].append(
                    {"image_path": file, "pre_description": pre_description}
                )
        return [image_metadata]
    else:
        return load_config(task_path)



import json
import os
from typing import Any, Dict, List, Optional

from cicada.core.utils import get_image_paths, image_to_base64, is_base64_encoded


def _create_text_content(text: str) -> dict:
    """Create text content message"""
    return {"type": "text", "text": text}


def _create_image_content(image_data: str) -> dict:
    """Create image content message from base64 encoded image string"""
    assert is_base64_encoded(image_data), "image_data must be base64 encoded"

    return {
        "type": "image_url",
        "image_url": {"url": f"data:image/jpeg;base64,{image_data}"},
    }


class PromptBuilder:
    """A utility class for constructing prompts with text and images.

    This class is designed to build a list of messages that can be used as input for
    models that accept multi-modal prompts (e.g., text and images). Messages can include
    system prompts, user prompts with text, and user prompts with images.

    Attributes:
        messages (list): A list of messages, where each message is a dictionary
            containing a role ("system" or "user") and content (text or image data).
    """

    def __init__(self):
        """Initialize the PromptBuilder with an empty list of messages."""
        self.messages = []
        self.tools = None  # Add an attribute to hold tools

    def add_system_message(self, content):
        """Add a system prompt to the messages.

        Args:
            content (str): The content of the system prompt.
        """
        self.messages.append({"role": "system", "content": content})

    def add_user_message(self, content):
        """Add a user prompt with text content to the messages.

        Args:
            content (str): The text content of the user prompt.
        """
        self.add_text(content)

    def add_images(
        self, image_data: list[str] | str, msg_index: Optional[int] = None
    ) -> int:
        """Add images to the messages. If msg_index is provided, the images will be appended to the existing message at that index to form a multi-content message.

        Accepts a list of image paths or a single image path. Each image is converted
        to a base64-encoded string and added as a user message with image content.

        Args:
            image_data (list[str] | str): A list of image paths or a single image path.
            msg_index (Optional[int]): The index of the message in the messages list. Useful when appending to an existing message.

        Returns:
            msg_index (int): The index of the message in the messages list, or -1 if no valid images were found.
        """

        image_files = get_image_paths(image_data)
        if not image_files:
            return -1  # No valid images found

        # Convert images to base64 and create image content
        new_content = [
            _create_image_content(image_to_base64(image_file))
            for image_file in image_files
        ]

        if msg_index is None:
            self.messages.append({"role": "user", "content": new_content})
            return len(self.messages) - 1

        existing_content = self.messages[msg_index]["content"]

        if isinstance(existing_content, str):
            # Convert single text content to a multi-content message
            self.messages[msg_index]["content"] = [
                _create_text_content(existing_content),
                *new_content,
            ]
        elif isinstance(existing_content, list):
            # Append to existing multi-content message
            self.messages[msg_index]["content"].extend(new_content)

        return msg_index

    def add_text(self, content: str, msg_index: Optional[int] = None) -> int:
        """Add a user message with text content to the messages. If msg_index is provided, the text will be appended to the existing message at that index to form a multi-content message.

        Args:
            content (str): The text content of the user message.
            msg_index (Optional[int]): The index of the message in the messages list. Useful when appending to an existing message.

        Return:
            msg_index (int): The index of the message in the messages list.
        """

        if msg_index is None:
            self.messages.append({"role": "user", "content": content})
            return len(self.messages) - 1

        existing_content = self.messages[msg_index]["content"]
        new_content = _create_text_content(content)

        if isinstance(existing_content, str):
            # Convert single text content to a multi-content message
            self.messages[msg_index]["content"] = [
                _create_text_content(existing_content),
                new_content,
            ]
        elif isinstance(existing_content, list):
            # Append to existing multi-content message
            self.messages[msg_index]["content"].append(new_content)

        return msg_index


class DesignGoal:
    """Represents a design goal, which can be defined by either text, images, or both.

    A design goal encapsulates the user's input, which can be in the form of a textual
    description, one or more images, or a combination of both. Images can be provided
    as paths to individual image files or as a path to a folder containing multiple images.

    Args:
        text (Optional[str]): A textual description of the design goal. Defaults to None.
        images (Optional[list[str]]): A list of image file paths or a single folder path
            containing images. Defaults to None.
        extra (Optional[Dict[str, Any]]): Additional information related to the design goal,
            such as original user input or decomposed part list, etc. Defaults to an empty dictionary.

    Raises:
        ValueError: If neither `text` nor `images` is provided.

    Attributes:
        text (Optional[str]): The textual description of the design goal.
        images (Optional[list[str]]): A list of image file paths or a single folder path.
        extra (Dict[str, Any]): Additional information related to the design goal, such as
            original user input or decomposed part list, etc.
    """

    def __init__(
        self,
        text: Optional[str] = None,
        images: Optional[List[str]] = None,
        extra: Optional[Dict[str, Any]] = None,
    ):
        # Validate that at least one of text or images is provided
        if text is None and images is None:
            raise ValueError("Either 'text' or 'images' must be provided.")

        self.text = text
        self.images = images
        # extra information, such as original user input, decomposed part list etc.
        self.extra = extra if extra else {}

    def __str__(self):
        return (
            f"DesignGoal(text='{self.text}', images={self.images}, extra={self.extra})"
        )

    def __repr__(self):
        return self.__str__()

    def to_dict(self) -> Dict[str, Any]:
        """Convert the DesignGoal object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the DesignGoal object.
        """
        return {
            "text": self.text,
            "images": self.images,
            "extra": self.extra,
        }

    def to_json(self) -> str:
        """Convert the DesignGoal object to a JSON string.

        Returns:
            str: A JSON string representation of the DesignGoal object.
        """
        return json.dumps(self.to_dict(), indent=4)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DesignGoal":
        """Create a DesignGoal object from a dictionary.

        Args:
            data (Dict[str, Any]): A dictionary containing the design goal data.

        Returns:
            DesignGoal: A DesignGoal object.
        """
        return cls(
            text=data.get("text"),
            images=data.get("images"),
            extra=data.get("extra"),
        )

    @classmethod
    def from_json(cls, json_str: str) -> "DesignGoal":
        """Create a DesignGoal object from a JSON string.

        Args:
            json_str (str): A JSON string containing the design goal data.

        Returns:
            DesignGoal: A DesignGoal object.
        """
        if os.path.isfile(json_str):
            with open(json_str, "r") as f:
                data = json.load(f)
        else:
            data = json.loads(json_str)
        return cls.from_dict(data)



import argparse
import logging
import os
import tkinter as tk
from tkinter import simpledialog
from typing import List, Literal

from trimesh import load_mesh

from cicada.coding.coder import Coder
from cicada.core.basics import DesignGoal
from cicada.core.utils import (
    colorstring,
    cprint,
    find_files_with_extensions,
    load_config,
    load_prompts,
    setup_logging,
)
from cicada.describe.describer_v2 import Describer
from cicada.feedback.feedback_judge import FeedbackJudge
from cicada.feedback.visual_feedback import VisualFeedback
from cicada.geometry_pipeline.snapshots import (
    generate_snapshots,
    preview_mesh_interactively,
)

logger = logging.getLogger(__name__)


class CodeExecutionLoop:
    def __init__(
        self,
        describer: Describer,
        coder: Coder,
        visual_feedback: VisualFeedback,
        feedback_judge: FeedbackJudge,
        max_design_iterations=10,
        max_coding_iterations=5,
    ):
        self.describer = describer
        self.coder = coder
        self.visual_feedback = visual_feedback
        self.feedback_judge = feedback_judge
        self.max_design_iterations = max_design_iterations
        self.max_coding_iterations = max_coding_iterations

    def run(
        self,
        design_goal: DesignGoal,
        output_dir: str,
        max_design_iterations: int = 10,
        stop_threshold: float = 0.8,
    ):
        # step 0: prepare output directory
        os.makedirs(output_dir, exist_ok=True)

        # Save initial design goal and refined design goal to output directory
        self._save_design_goal(
            design_goal,
            os.path.join(output_dir, "initial_design_goal.json"),
        )

        # Step 1: Refine the design goal using the Describer
        logger.info("START [refine_design_goal]")
        refined_design_goal = self._refine_design_goal(design_goal)
        logger.info("DONE [refine_design_goal]")
        self._save_design_goal(
            refined_design_goal,
            os.path.join(output_dir, "refined_design_goal.json"),
        )

        iteration = 0
        best_code = None
        best_feedbacks = None
        best_coding_plan = None
        is_completed = False

        # Step 2: Proceed with the code generation and execution loop using the refined design goal
        while iteration < max_design_iterations:
            iteration_dir = os.path.join(output_dir, f"iteration_{iteration + 1}")
            os.makedirs(iteration_dir, exist_ok=True)

            # Generate executable code using the Coder class
            logger.info("START [generate_executable_code]")
            generated_code, coding_plan = self.coder.generate_executable_code(
                refined_design_goal,
                feedbacks=best_feedbacks,
                generated_code=best_code,
                coding_plan=best_coding_plan,
            )
            logger.info("DONE [generate_executable_code]")
            if generated_code is None:
                logger.error(
                    f"Iteration {iteration + 1} - No executable code generated."
                )
                iteration += 1
                continue

            # Save generated code
            self.coder.code_generator.save_code_to_file(
                generated_code, os.path.join(iteration_dir, "code.py")
            )

            # render_from_code
            logger.info("START [render_from_code]")
            is_success, messages, render_dir = self.coder.render_from_code(
                generated_code, iteration_dir, format="stl"
            )
            logger.info("DONE [render_from_code]")
            if not is_success:
                logger.error(
                    f"Iteration {iteration + 1} - Rendering failed: {messages}"
                )
                iteration += 1
                continue
            else:
                human_feedback = self._preview_mesh(render_dir)

            # get_visual_feedback
            logger.info("START [get_visual_feedback]")
            is_success, visual_feedback = self._get_visual_feedback(
                refined_design_goal, render_dir, snapshot_directions="omni"
            )
            logger.info("DONE [get_visual_feedback]")
            if not is_success:
                logger.error(
                    colorstring(
                        f"Iteration {iteration + 1} - Visual feedback failed", "red"
                    )
                )
                iteration += 1
                continue

            # Save visual feedback
            with open(os.path.join(iteration_dir, "visual_feedback.txt"), "w") as f:
                f.write(visual_feedback)

            # Update best feedback and code based on feedback evaluation
            logger.info("START [update_best_feedback_and_code]")
            if best_feedbacks is None:
                best_code = generated_code
                best_feedbacks = visual_feedback
                logger.info(
                    colorstring(
                        f"Iteration {iteration + 1} - Initial feedback received",
                        "cyan",
                    )
                )

            # Check if the design goal has been achieved
            logger.info("START [check_design_goal]")
            is_achieved, score = self.feedback_judge.is_design_goal_achieved(
                visual_feedback, refined_design_goal.text
            )
            logger.info("DONE [check_design_goal]")
            if is_achieved or score >= stop_threshold:
                best_code = generated_code
                best_feedbacks = visual_feedback
                best_coding_plan = coding_plan
                logger.info(
                    colorstring(
                        f"Iteration {iteration + 1} - Design goal achieved! (Score: {score})",
                        "white",
                    )
                )
                is_completed = True
                break

            iteration += 1

        if is_completed:
            finish_message = (
                f"SUCCESS!"
                f"Design task completed after {iteration} iterations."
                f"Best code: {best_code}"
                f"Best feedbacks: {best_feedbacks}"
                f"Best coding plan: {best_coding_plan}"
            )
            msg_color = "bright_green"
        else:
            finish_message = f"Design task not completed after {iteration} iterations."
            msg_color = "bright_red"

        logger.info(colorstring(finish_message, msg_color))
        return best_code, best_feedbacks, best_coding_plan

    def _refine_design_goal(self, design_goal: DesignGoal) -> DesignGoal:
        """
        Refines the design goal using the Describer.

        Args:
            design_goal (DesignGoal): The original design goal to be refined.

        Returns:
            DesignGoal: The refined design goal.
        """
        # Use the Describer to refine the design goal
        refined_design_goal = self.describer.design_feedback_loop(design_goal)

        return refined_design_goal

    def _get_visual_feedback(
        self,
        design_goal: DesignGoal,
        render_dir: str,
        snapshot_directions: str | Literal["common", "box", "omni"] = "common",
    ) -> tuple[bool, str]:
        """
        find out rendered object inside render_dir, then take snapshots according to snapshot_directions, then compare with design_goal and generate feedbacks
        """
        # prefer stl if available, otherwise obj, otherwise step. return only one
        rendered_obj_path = find_files_with_extensions(
            render_dir, ["stl", "step", "obj"], return_all=False
        )

        if rendered_obj_path is None:
            logger.error("No rendered object found in the render path.")
            return False, "No rendered object found in the render path."

        # take snapshots
        snapshots_dir = os.path.join(render_dir, "snapshots")
        os.makedirs(snapshots_dir, exist_ok=True)  # Ensure the directory exists
        snapshot_paths = generate_snapshots(
            file_path=rendered_obj_path,
            output_dir=snapshots_dir,
            direction=snapshot_directions,
            mesh_color=[0, 102, 204],
            reaxis_gravity=True,
        )
        logger.info(colorstring(f"Snapshot paths: {snapshot_paths}", "magenta"))

        # compare with design_goal and generate feedbacks
        visual_feedback = self.visual_feedback.generate_feedback_paragraph(
            design_goal.text, design_goal.images, snapshot_paths
        )

        logger.info(colorstring(f"Visual feedback: {visual_feedback}", "white"))

        return True, visual_feedback

    def _preview_mesh(self, render_dir: str) -> str | None:
        """
        Preview the rendered mesh interactively and collect human feedback.

        Args:
            render_dir (str): The directory containing the rendered mesh file.

        Returns:
            str | None: The human feedback provided by the user as a string.
                    Returns `None` if no feedback is provided or if no rendered object is found.
        """
        # Find the rendered mesh file
        rendered_obj_path = find_files_with_extensions(
            render_dir, ["stl", "step", "obj"], return_all=False
        )

        if rendered_obj_path is None:
            logger.error("No rendered object found in the render path.")
            return None

        # Load the mesh
        mesh = load_mesh(rendered_obj_path)

        # Preview the mesh interactively
        preview_mesh_interactively(
            mesh,
            direction="front",
            reaxis_gravity=True,
            mesh_color=[0, 102, 204],  # Example color
        )

        # Create a simple GUI to collect human feedback
        root = tk.Tk()
        root.withdraw()  # Hide the root window

        # Show an input dialog to collect feedback
        feedback = simpledialog.askstring(
            "Feedback", "Please provide your feedback on the mesh:"
        )

        if feedback:
            logger.info(f"Human feedback received: {feedback}")
            # Process the feedback as needed
            feedback_file_path = os.path.join(render_dir, "human_feedback.txt")
            with open(feedback_file_path, "w") as f:
                f.write(feedback)
            logger.info(f"Feedback saved to {feedback_file_path}")
        else:
            logger.info("No feedback provided.")

        root.destroy()  # Close the GUI

        return feedback

    def _save_design_goal(self, design_goal: DesignGoal, file_path: str):
        """
        Save the design goal (text and reference images) as a JSON file.

        Args:
            design_goal (DesignGoal): The design goal to save.
            file_path (str): The path to save the JSON file.
        """
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        try:
            with open(file_path, "w") as f:
                f.write(design_goal.to_json())
            logger.info(f"Design goal saved to {file_path}")
        except Exception as e:
            logger.error(f"Failed to save design goal to {file_path}: {e}")
            cprint(design_goal, "bright_yellow")
            raise e


def parse_args():
    """
    Parse command-line arguments.

    Returns:
        argparse.Namespace: Parsed arguments.
    """
    parser = argparse.ArgumentParser(description="Assistive Large Language Model")
    parser.add_argument(
        "--config",
        default="config",
        help="Path to the configuration YAML folder",
    )
    parser.add_argument(
        "--prompts",
        default="prompts",
        help="Path to the prompts YAML folder",
    )
    parser.add_argument(
        "design_task",
        help="Description of the design task (e.g., 'create a simple skateboard')",
    )
    parser.add_argument(
        "-img",
        "--ref_images",
        type=str,
        default=None,
        help="Paths to reference images for the design",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        default="./design_task_results",
        help="Directory to save the results of the design task",
    )
    return parser.parse_args()


def init_models(config_path: str, prompts_path: str):
    """
    Initialize all models and components required for the code execution loop.

    Args:
        config_path (str): Path to the configuration YAML file or folder.
        prompts_path (str): Path to the prompts YAML file or folder.

    Returns:
        tuple: A tuple containing the initialized components:
            - describer: Describer instance
            - coder: Coder instance
            - visual_feedback: VisualFeedback instance
            - feedback_judge: FeedbackJudge instance
    """
    # ===== describer agent =====
    describer_config = load_config(config_path, "describe-vlm")
    describer = Describer(
        describer_config["api_key"],
        describer_config.get("api_base_url"),
        describer_config.get("model_name"),
        describer_config.get("org_id"),
        load_prompts(prompts_path, "describe-vlm"),
        **describer_config.get("model_kwargs", {}),
    )

    # ===== coder agent =====
    coder_config = load_config(config_path, "code-llm")

    try:
        master_code_llm_config = load_config(config_path, "master-code-llm")
    except FileNotFoundError as e:
        master_code_llm_config = None
        logger.warning(f"Master code LLM configuration not found: {e}")

    coder = Coder(
        coder_config.get("api_key"),
        coder_config.get("api_base_url"),
        coder_config.get("model_name"),
        coder_config.get("org_id"),
        load_prompts(prompts_path, "code-llm"),
        code_master_config=master_code_llm_config,
        **coder_config.get("model_kwargs", {}),
    )

    # ===== visual feedback =====
    visual_feedback_config = load_config(config_path, "visual_feedback")
    visual_feedback = VisualFeedback(
        visual_feedback_config["api_key"],
        visual_feedback_config.get("api_base_url"),
        visual_feedback_config.get("model_name"),
        visual_feedback_config.get("org_id"),
        load_prompts(prompts_path, "visual_feedback"),
        **visual_feedback_config.get("model_kwargs", {}),
    )

    # ===== feedback judge =====
    feedback_judge_config = load_config(config_path, "feedback_judge")
    feedback_judge = FeedbackJudge(
        feedback_judge_config["api_key"],
        feedback_judge_config.get("api_base_url"),
        feedback_judge_config.get("model_name"),
        feedback_judge_config.get("org_id"),
        load_prompts(prompts_path, "feedback_judge"),
        **feedback_judge_config.get("model_kwargs", {}),
    )

    return describer, coder, visual_feedback, feedback_judge


def main():
    """
    Main function to run the code execution loop.
    """
    # Parse command-line arguments
    args = parse_args()

    # Initialize models
    describer, coder, visual_feedback, feedback_judge = init_models(
        args.config, args.prompts
    )

    # ===== code execution loop =====
    code_execution_loop = CodeExecutionLoop(
        describer=describer,
        coder=coder,
        visual_feedback=visual_feedback,
        feedback_judge=feedback_judge,
    )

    # Create the design goal
    design_goal = DesignGoal(args.design_task, args.ref_images)

    # Use the output directory specified in the command-line arguments
    output_dir = args.output_dir

    # Run the code execution loop
    code_execution_loop.run(design_goal, output_dir)


if __name__ == "__main__":
    setup_logging()
    main()



import logging
from typing import Literal

import httpx

# Set up logging
logger = logging.getLogger(__name__)


def fetch_content(
    url,
    return_format: Literal["text", "markdown", "html", "screenshot"] = "markdown",
    disable_gfm: Literal[False, True, "table"] = False,
    bypass_cache=False,
    with_generated_alt=False,
    remove_images=False,
    timeout=None,
    json_response=True,
    with_links_summary=False,
    with_images_summary=False,
):
    """
    Fetch content from a specified URL using the r.jina.ai API.

    :param url: The URL to fetch content from.
    :param return_format: The format of the returned content, options are 'text', 'markdown', 'html', 'screenshot'.
    :param json_response: Boolean flag to toggle JSON response format.
    :param bypass_cache: Boolean flag to bypass cache.
    :param disable_gfm: Boolean flag to disable Github Flavored Markdown or 'table' mode.
    :param remove_images: Boolean flag to remove all images from the response.
    :param timeout: Timeout in seconds for waiting the webpage to load.
    :param with_generated_alt: Boolean flag to include generated alt text for images.
    :param with_links_summary: Boolean flag to include a summary of all links at the end.
    :param with_images_summary: Boolean flag to include a summary of all images at the end.
    :return: The fetched content in the specified format.
    """
    api_url = f"https://r.jina.ai/{url}"
    headers = {"X-Return-Format": return_format}

    if bypass_cache:
        headers["X-No-Cache"] = "true"

    if remove_images:
        headers["X-Retain-Images"] = "none"

    if disable_gfm is not False:
        headers["X-No-Gfm"] = "true" if disable_gfm == True else "table"

    if timeout is not None:
        headers["X-Timeout"] = str(timeout)

    if with_generated_alt:
        headers["X-With-Generated-Alt"] = "true"

    if with_links_summary:
        headers["X-With-Links-Summary"] = "true"

    if with_images_summary:
        headers["X-With-Images-Summary"] = "true"

    if json_response:
        headers["Accept"] = "application/json"

    try:
        response = httpx.get(api_url, headers=headers, timeout=10)
        response.raise_for_status()
        content = response.text
        logger.info(
            f"Successfully fetched content from: {url} in {return_format} format"
        )
    except httpx.RequestError as e:
        content = f"error: {str(e)}"
        logger.error(f"Failed to fetch content from {url}. Error: {str(e)}")
    return content


# Example usage:
if __name__ == "__main__":

    from cicada.core.utils import setup_logging

    setup_logging()

    url = "https://thgilkao.github.io/picx-images-hosting/IMG_20230913_211233.jpg"  # @param {type:"string"}
    content_format = "markdown"  # @param ["text", "markdown", "html", "screenshot"]
    json_response = False  # @param {type:"boolean"}
    bypass_cache = False  # @param {type:"boolean"}
    disable_gfm = True  # @param {type:"boolean"}
    remove_images = False  # @param {type:"boolean"}
    timeout = 60  # @param {type:"integer"}
    with_generated_alt = True  # @param {type:"boolean"}
    with_links_summary = True  # @param {type:"boolean"}
    with_images_summary = True  # @param {type:"boolean"}

    fetched_content = fetch_content(
        url,
        return_format=content_format,
        json_response=json_response,
        bypass_cache=bypass_cache,
        disable_gfm=disable_gfm,
        remove_images=remove_images,
        timeout=timeout,
        with_generated_alt=with_generated_alt,
        with_links_summary=with_links_summary,
        with_images_summary=with_images_summary,
    )
    logger.info("\nFetched Content:\n")
    logger.info(fetched_content)



import importlib
import inspect
import logging
from functools import lru_cache

from cicada.core.utils import colorstring

logger = logging.getLogger(__name__)


class CodeDocHelper:
    def __init__(self):
        """
        Initialize the CodeDocHelper
        """
        pass

    def get_function_info(self, function_path, with_docstring=False):
        """
        Retrieve the signature and optionally the docstring of a function or class method.

        Parameters:
        function_path (str): The full import path to the function (e.g., "build123d.Box.bounding_box").
        with_docstring (bool): If True, include the docstring in the output.

        Returns:
        dict: A dictionary containing the function name, signature, and docstring (optional).
        """
        logger.debug(
            colorstring(
                f"Getting function info for {function_path}", color="bright_blue"
            )
        )

        try:
            parts = function_path.split(".")
            module_name = parts[0]
            module = importlib.import_module(module_name)

            # Traverse the path to get the function or method
            obj = module
            for part in parts[1:]:
                obj = getattr(obj, part)

            # Check if the object is a function or method
            if inspect.isroutine(obj):
                signature = str(inspect.signature(obj))
                if inspect.ismethod(obj) and signature.startswith("(self, "):
                    signature = f"({signature[7:]}"  # Remove "(self, "

                # # `parts[-1]` version: clearer to read, but may be ambiguous for llm
                # data = {"name": parts[-1], "signature": f"{parts[-1]}{signature}"}

                # `function_path` version: more verbose, but provide which module it comes from
                data = {
                    "name": function_path,
                    "signature": f"{function_path}{signature}",
                }
                if with_docstring:
                    data["docstring"] = inspect.getdoc(obj) or "No docstring available."
                return data

            return {"error": f"Object '{function_path}' is not a function or method."}
        except Exception as e:
            return {"error": f"Error getting function info: {e}"}

    @lru_cache(maxsize=128)
    def get_class_info(self, class_path, with_docstring=False):
        """
        Retrieve the signature, methods, variables, and optionally the docstring of a class.

        Parameters:
        class_path (str): The full import path to the class (e.g., "build123d.Box").
        with_docstring (bool): If True, include the docstring in the output.

        Returns:
        dict: A dictionary containing the class name, signature, flattened methods, variables, and docstring (optional).
        """
        logger.debug(colorstring(f"Getting class info for {class_path}", "cyan"))

        try:
            parts = class_path.split(".")
            module_name = parts[0]
            module = importlib.import_module(module_name)

            # Traverse the path to get the class
            cls = module
            for part in parts[1:]:
                cls = getattr(cls, part)

            # Handle __init__ method
            init_func = cls.__init__ if "__init__" in cls.__dict__ else None
            signature = str(inspect.signature(init_func)) if init_func else "()"
            if signature.startswith("(self, "):
                signature = signature.replace("(self, ", "(")

            data = {
                "name": class_path,
                "signature": f"{class_path}{signature}",
                "methods": [],  # Flattened methods (as standalone functions)
                "variables": [],
            }
            if with_docstring:
                data["docstring"] = inspect.getdoc(cls) or "No docstring available."

            # Collect methods and flatten them using get_function_info
            for name, member in inspect.getmembers(cls):
                if inspect.isroutine(member) and not name.startswith("_"):
                    method_path = f"{class_path}.{name}"
                    method_info = self.get_function_info(method_path, with_docstring)
                    if "error" not in method_info:
                        data["methods"].append(method_info)

            # Collect variables
            for name, member in inspect.getmembers(cls):
                if not inspect.isroutine(member) and not name.startswith("_"):
                    data["variables"].append(name)

            return data
        except Exception as e:
            return {"error": f"Error getting class info: {e}"}

    def get_module_info(self, module_name, with_docstring=False):
        """
        Retrieve the classes, functions, variables, and optionally the docstring of a module.

        Parameters:
        module_name (str): The name of the module.
        with_docstring (bool): If True, include the docstring in the output.

        Returns:
        dict: A dictionary containing the module name, classes, functions, variables, and docstring (optional).
        """
        logger.debug(colorstring(f"Getting module info for {module_name}", "yellow"))

        try:
            module = importlib.import_module(module_name)
            data = {
                "name": module_name,
                "classes": [],
                "functions": [],
                "variables": [],
            }
            if with_docstring:
                data["docstring"] = inspect.getdoc(module) or "No docstring available."

            # Helper function to check if member belongs to the module
            def belongs_to_module(member):
                return hasattr(member, "__module__") and member.__module__.startswith(
                    module_name
                )

            # Collect classes
            for name, member in inspect.getmembers(module, inspect.isclass):
                if belongs_to_module(member):
                    class_info = self.get_class_info(
                        f"{module_name}.{name}", with_docstring
                    )
                    if "error" not in class_info:
                        data["classes"].append(class_info)

            # Collect functions (including built-in functions)
            for name, member in inspect.getmembers(
                module, lambda m: inspect.isroutine(m)
            ):
                if belongs_to_module(member):
                    func_info = self.get_function_info(
                        f"{module_name}.{name}", with_docstring
                    )
                    if "error" not in func_info:
                        data["functions"].append(func_info)

            # Collect variables
            for name, member in inspect.getmembers(module):
                if (
                    not inspect.isclass(member)
                    and not inspect.isroutine(member)
                    and not name.startswith("_")
                    and belongs_to_module(member)
                ):
                    variable_info = {
                        "name": name,
                        "value": str(member),
                        "type": type(member).__name__,
                    }
                    data["variables"].append(variable_info)

            return data
        except ImportError:
            return {"error": f"Module '{module_name}' not found."}
        except Exception as e:
            return {"error": f"An unexpected error occurred: {str(e)}"}

    def get_info(self, path, with_docstring=True):
        """
        Retrieve information about a module, class, function, or variable.
        """

        try:
            parts = path.split(".")
            module_name = parts[0]
            logger.debug(f"Importing module: {module_name}")
            module = importlib.import_module(module_name)
            logger.debug(f"Successfully imported module: {module_name}")

            # Traverse the path to get the member
            obj = module
            for part in parts[1:]:
                logger.debug(f"Traversing path: {part}")
                obj = getattr(obj, part)
                logger.debug(f"Retrieved object: {obj}")

            logger.debug(f"Inspecting object: {obj}")
            # Determine the type of the member
            if (
                inspect.isfunction(obj)
                or inspect.isbuiltin(obj)
                or inspect.ismethod(obj)
            ):
                logger.debug(colorstring(f"Detected function: {path}", "magenta"))
                result = self.get_function_info(path, with_docstring=with_docstring)
            elif inspect.isclass(obj):
                logger.debug(colorstring(f"Detected class: {path}", "cyan"))
                result = self.get_class_info(path, with_docstring=with_docstring)
            elif inspect.ismodule(obj):
                logger.debug(colorstring(f"Detected module: {path}", "yellow"))
                result = self.get_module_info(path, with_docstring=with_docstring)
            else:
                logger.debug(f"Detected variable: {path}")
                # Handle variables or other types
                result = {
                    "name": path,
                    "type": type(obj).__name__,
                    "value": str(obj),
                }

            # Save debug info
            logger.debug(f"Saving debug info to debug_info.json")
            if logging.root.level == logging.DEBUG:
                import json

                with open("debug_info.json", "w") as f:
                    json.dump(result, f, indent=4)

            return result
        except ImportError as e:
            logger.error(f"ImportError: {e}")
            return {"error": f"Module '{module_name}' not found."}
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return {
                "error": f"Member '{parts[-1]}' not found in module '{module_name}'."
            }
        except Exception as e:
            logger.error(f"Unexpected error: {e}")
            return {"error": f"An unexpected error occurred: {str(e)}"}

    def dict_to_markdown(self, data, show_docstring=True):
        """
        Convert a dictionary containing module, class, function, or variable information into a Markdown-formatted string.

        Parameters:
        data (dict): A dictionary containing module, class, function, or variable information.
        show_docstring (bool): If True, include docstrings in the output.

        Returns:
        str: A Markdown-formatted string.
        """
        if "error" in data:
            return f"# Error\n{data['error']}"
        if "classes" in data and "functions" in data:
            # It's module info
            markdown_output = f"# Module: {data['name']}\n\n"
            if show_docstring and "docstring" in data:
                markdown_output += (
                    f"- **Docstring**:\n```markdown\n{data['docstring']}\n```\n\n"
                )
            if data["classes"]:
                markdown_output += "## Classes\n"
                for cls in data["classes"]:
                    markdown_output += f"### {cls['name']}\n"
                    markdown_output += f"- **Signature**: `{cls['signature']}`\n"
                    if show_docstring and "docstring" in cls:
                        markdown_output += f"- **Docstring**:\n```markdown\n{cls['docstring']}\n```\n\n"
            if data["functions"]:
                markdown_output += "## Functions\n"
                for func in data["functions"]:
                    markdown_output += f"### {func['name']}\n"
                    markdown_output += f"- **Signature**: `{func['signature']}`\n"
                    if show_docstring and "docstring" in func:
                        markdown_output += f"- **Docstring**:\n```markdown\n{func['docstring']}\n```\n\n"
            if data["variables"]:
                markdown_output += "## Variables\n"
                for var in data["variables"]:
                    markdown_output += (
                        f"- **{var['name']}** (`{var['type']}`): {var['value']}\n"
                    )
            return markdown_output
        elif "methods" in data:
            # It's class info
            markdown_output = f"# Class: {data['name']}\n\n"
            markdown_output += f"- **Signature**: `{data['signature']}`\n"
            if show_docstring and "docstring" in data:
                markdown_output += (
                    f"- **Docstring**:\n```markdown\n{data['docstring']}\n```\n\n"
                )
            markdown_output += "## Methods\n"
            for method in data["methods"]:
                markdown_output += f"### {method['name']}\n"
                markdown_output += f"- **Signature**: `{method['signature']}`\n"
                if show_docstring and "docstring" in method:
                    markdown_output += (
                        f"- **Docstring**:\n```markdown\n{method['docstring']}\n```\n\n"
                    )
            markdown_output += "## Variables\n"
            for var in data["variables"]:
                markdown_output += f"- {var}\n"
            return markdown_output
        elif "signature" in data:
            # It's function info
            markdown_output = f"# Function: {data['name']}\n\n"
            markdown_output += f"- **Signature**: `{data['signature']}`\n"
            if show_docstring and "docstring" in data:
                markdown_output += (
                    f"- **Docstring**:\n```markdown\n{data['docstring']}\n```\n"
                )
            return markdown_output
        elif "type" in data:
            # It's a variable
            markdown_output = f"# Variable: {data['name']}\n\n"
            markdown_output += f"- **Type**: `{data['type']}`\n"
            markdown_output += f"- **Value**: `{data['value']}`\n"
            return markdown_output
        else:
            return "# Unknown\nInvalid data format."


def doc_helper(import_path, with_docstring: bool = False):
    """
    Query module, class, function, or method information and return it in Markdown format.
    """
    _helper = CodeDocHelper()
    info = _helper.get_info(import_path, with_docstring=with_docstring)
    markdown_formatted_str = _helper.dict_to_markdown(
        info, show_docstring=with_docstring
    )
    return markdown_formatted_str


if __name__ == "__main__":

    import argparse

    from cicada.core.utils import colorstring, setup_logging  # For fuzzy matching

    parser = argparse.ArgumentParser(
        description="Query module, class, function, or method information."
    )
    parser.add_argument(
        "path",
        type=str,
        default="math.sqrt",
        nargs="?",  # Makes the path argument optional (consumes one argument if provided)
        help="The full import path or keyword to query (e.g., 'Box' or 'build123d.Box').",
    )
    parser.add_argument(
        "--docstring",
        action="store_true",
        help="Include the docstring in the output (only applies to path queries).",
    )
    parser.add_argument(
        "--no_output",
        action="store_true",
        help="Disable output to the console.",
    )
    parser.add_argument("--debug", action="store_true", help="Enable debug logging.")
    args = parser.parse_args()

    setup_logging(
        log_level="DEBUG" if args.debug else "INFO",
        log_file="/tmp/cicada/code_dochelper.log",
    )

    # Initialize the CodeDocHelper
    helper = CodeDocHelper()

    # Get the information based on the provided path
    info = helper.get_info(args.path, with_docstring=args.docstring)

    # Convert the information to Markdown and print it
    markdown_formatted_str = helper.dict_to_markdown(
        info, show_docstring=args.docstring
    )
    print(markdown_formatted_str)
    # Save debug info

    logger.debug(f"Saving debug info to debug_info.json")
    if logging.root.level == logging.DEBUG:
        # Save the output to a file
        with open("debug_info.md", "w") as f:
            f.write(markdown_formatted_str)



import argparse
import logging
import os
import random
import sys
from typing import List

from tqdm import tqdm

_current_dir = os.path.dirname(os.path.abspath(__file__))
_parent_dir = os.path.dirname(_current_dir)
sys.path.extend([_current_dir, _parent_dir])

from common import vlm
from common.utils import colorstring, image_to_base64, load_config, load_prompts
from describe.utils import load_object_metadata, save_descriptions

# Configure logging
logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(levelname)s - %(message)s"
)

MAX_IMAGES_PER_QUERY = 4  # to prevent input exceed max input token


class DescriberVLM(vlm.VisionLanguageModel):
    def __init__(
        self,
        api_key,
        api_base_url,
        model_name,
        org_id,
        prompt_templates,
        **model_kwargs,
    ):
        super().__init__(
            api_key,
            api_base_url,
            model_name,
            org_id,
            **model_kwargs,
        )
        self.reverse_engineer_prompt = prompt_templates.get("reverse_engineer", {})
        self.featurize_design_prompt = prompt_templates.get("featurize_design", {})

    def generate_descriptions_metadata(self, objects: List[dict], save=True):
        """
        Generate descriptions for the given objects and their images.

        :param objects: List of objects with their metadata.
        :return: List of generated descriptions of the images.
        """
        description_collection = []
        for obj in tqdm(objects, desc="Describing objects", unit="object", leave=True):
            logging.debug(colorstring(f"Processing object: {obj['object_id']}", "cyan"))
            object_id = obj["object_id"]
            object_description = obj.get("object_description", "")

            # Load images in batch
            image_paths = [
                os.path.join(obj["base_path"], img["image_path"])
                for img in obj["images"]
            ]
            pre_descriptions = [img.get("pre_description", "") for img in obj["images"]]

            # open and convert image to base64
            image_data = [image_to_base64(image_path) for image_path in image_paths]

            # Generate description for each image
            metadata_dict = {
                "object_id": object_id,
                "image_path": image_paths,
                "pre_description": pre_descriptions,
                "generated_description": None,
                "error": None,
            }
            try:
                # Flatten the sampled pairs into a list
                # images_with_pre_descriptions = [
                #     item for pair in zip(pre_descriptions, image_data) for item in pair
                # ]
                # sample uniformly half of the zip
                packed_images = list(
                    zip(pre_descriptions, image_data)
                )  # Pack images and pre-descriptions together
                images_with_pre_descriptions = random.sample(
                    packed_images, min(MAX_IMAGES_PER_QUERY, len(packed_images))
                )
                images_with_pre_descriptions = [
                    item for pair in images_with_pre_descriptions for item in pair
                ]  # Flatten the sampled pairs into a list
                description = self.query_with_image(
                    prompt=self.reverse_engineer_prompt["user_prompt_template"],
                    images_with_text=images_with_pre_descriptions,
                    system_prompt=self.reverse_engineer_prompt[
                        "system_prompt_template"
                    ],
                )

                metadata_dict.update({"generated_description": description})

            except Exception as e:
                metadata_dict.update({"error": str(e)})

            if save:
                save_descriptions(obj["base_path"], metadata_dict)
                logging.info(
                    colorstring(f"Saved description for object id: {object_id}", "cyan")
                )

            description_collection.append(metadata_dict)
            logging.info(
                colorstring(f"Generated description for object id: {object_id}", "cyan")
            )

        return description_collection


def _main():
    parser = argparse.ArgumentParser(description="Vision Language Model")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    parser.add_argument(
        "--prompts", default="prompts.yaml", help="Path to the prompts YAML file"
    )
    parser.add_argument(
        "--task",
        default="tasks.yaml",
        help="Path to the task YAML file or directory containing images from a single object",
    )
    parser.add_argument(
        "--save",
        action="store_true",
        help="Save generated descriptions to metadata.yaml",
    )
    args = parser.parse_args()

    config = load_config(args.config)
    image_metadata = load_object_metadata(args.task)

    describe_vlm_config = config["describe-vlm"]

    vlm = DescriberVLM(
        describe_vlm_config["api_key"],
        describe_vlm_config.get("api_base_url"),
        describe_vlm_config.get("model_name", "gpt-4o"),
        describe_vlm_config.get("org_id"),
        load_prompts(args.prompts, "describe-vlm"),
        **describe_vlm_config.get("model_kwargs", {}),
    )
    descriptions = vlm.generate_descriptions_metadata(image_metadata, args.save)

    for desc in descriptions:
        logging.debug(f"Object ID: {desc['object_id']}")
        logging.debug(f"Image Path: {desc['image_path']}")
        logging.debug(f"Pre-Description: {desc['pre_description']}")
        if "generated_description" in desc:
            logging.debug(f"Generated Description:\n{desc['generated_description']}")
        else:
            logging.debug(f"Error: {desc['error']}")
        logging.debug("-" * 40)


# Example usage
if __name__ == "__main__":
    _main()



import base64
import io
import json
import logging
import logging.config
import os
import re
from typing import Any, Dict, Iterable, List, Optional

import yaml
from blessed import Terminal
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from PIL import Image

logger = logging.getLogger(__name__)

# Initialize the terminal object
_term = Terminal()


def load_config(config_path: str, config_name: Optional[str] = None) -> dict:
    """Load a YAML configuration file or a specific configuration from a folder or file.

    Args:
        config_path (str): Path to the YAML configuration file or folder containing configuration files.
        config_name (Optional[str]): Name of the target configuration file (if config_path is a folder)
                                    or the key within the YAML file (if config_path is a file).
                                    If omitted and config_path is a file, the entire file is loaded.

    Returns:
        dict: Dictionary containing the configuration data.

    Raises:
        FileNotFoundError: If the specified `config_path` does not exist.
        yaml.YAMLError: If the YAML file is malformed or cannot be parsed.
        ValueError: If the `config_name` is not found in the configuration.
    """
    if os.path.isdir(config_path):
        # If config_path is a folder, config_name must be provided
        if config_name is None:
            raise ValueError(
                "config_name must be provided when config_path is a folder."
            )

        # Construct the full path to the config file
        config_file_path = os.path.join(config_path, f"{config_name}.yaml")
        if not os.path.exists(config_file_path):
            raise FileNotFoundError(
                f"Configuration file '{config_file_path}' not found in folder '{config_path}'."
            )

        with open(config_file_path, "r") as file:
            return yaml.safe_load(file)

    elif os.path.isfile(config_path):
        # If config_path is a file, load the YAML
        with open(config_path, "r") as file:
            config_data = yaml.safe_load(file)

        # If config_name is provided, extract the specific config
        if config_name is not None:
            if config_name not in config_data:
                raise ValueError(
                    f"Configuration key '{config_name}' not found in file '{config_path}'."
                )
            return config_data[config_name]

        # If config_name is omitted, return the entire config
        return config_data

    else:
        raise FileNotFoundError(f"Path '{config_path}' does not exist.")


def load_prompts(prompts_path: str, which_model: str) -> dict:
    """Load prompts from a YAML file and return prompts for a specific model.

    Args:
        prompts_path (str): Path to the YAML file containing prompts.
        which_model (str): Key specifying which model's prompts to load.

    Returns:
        dict: Dictionary containing prompts for the specified model.

    Raises:
        KeyError: If the specified `which_model` key is not found in the YAML file.
    """
    prompt_templates = load_config(prompts_path, which_model)
    return prompt_templates


def colorstring(
    message: Any,  # Accept any type of input
    color: Optional[str] = "green",
    bold: bool = False,
) -> str:
    """
    Returns a colored string using either ANSI escape codes or blessed terminal capabilities.

    :param message: The message to be colored. Can be of any type (e.g., str, int, float, bool).
    :param color: The color to apply. Supported colors: 'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white',
                 and their bright variants: 'bright_black', 'bright_red', 'bright_green', 'bright_yellow', 'bright_blue',
                 'bright_magenta', 'bright_cyan', 'bright_white'.
    :param bold: If True, applies bold styling to the text (only applicable when use_ansi=True).
    :return: A string with the specified color and styling.
    """
    color_mapping = {
        "black": _term.black,
        "blue": _term.blue,
        "cyan": _term.cyan,
        "green": _term.green,
        "magenta": _term.magenta,
        "red": _term.red,
        "white": _term.white,
        "yellow": _term.yellow,
        "bright_black": _term.bright_black,
        "bright_blue": _term.bright_blue,
        "bright_cyan": _term.bright_cyan,
        "bright_green": _term.bright_green,
        "bright_magenta": _term.bright_magenta,
        "bright_red": _term.bright_red,
        "bright_white": _term.bright_white,
        "bright_yellow": _term.bright_yellow,
    }

    # Convert the message to a string
    message_str = str(message)

    color_func = color_mapping.get(color.lower(), _term.white)
    styled_message = color_func(message_str)
    if bold:
        styled_message = _term.bold(styled_message)
    return styled_message


def cprint(message: Any, color: Optional[str] = "green", **kwargs) -> None:
    """
    Prints a colored string using blessed terminal capabilities.

    :param message: The message to be colored. Can be of any type (e.g., str, int, float, bool).
    :param color: The color to apply. Supported colors: 'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.
    :param kwargs: Additional keyword arguments to pass to the `print` function (e.g., `end`, `sep`, `file`, `flush`).
    """
    print(colorstring(message, color), **kwargs)


def get_image_paths(path: str | List[str]) -> List[str]:
    """
    Get image file paths from a specified folder, a single image file, or a list of image paths.

    Parameters:
    path (Union[str, List[str]]): The path to the folder, the single image file, or a list of image paths.

    Returns:
    List[str]: A list of image file paths.

    Raises:
    ValueError: If any path does not exist or is not a valid image file or folder of images.
    """
    valid_extensions = {".jpg", ".jpeg", ".png", ".bmp", ".gif"}

    def _get_single_image_path(p: str) -> List[str]:
        if not os.path.exists(p):
            raise ValueError(f"The path '{p}' does not exist.")

        if os.path.isfile(p):
            if os.path.splitext(p)[1].lower() in valid_extensions:
                return [p]
            raise ValueError(f"The file '{p}' is not a recognized image file.")

        if os.path.isdir(p):
            return [
                os.path.join(p, f)
                for f in os.listdir(p)
                if os.path.splitext(f)[1].lower() in valid_extensions
            ]

        raise ValueError(f"The path '{p}' is neither a file nor a directory of image.")

    if isinstance(path, str):
        return _get_single_image_path(path)
    elif isinstance(path, list):
        image_paths = []
        for p in path:
            image_paths.extend(_get_single_image_path(p))
        return image_paths
    else:
        raise ValueError("The input must be a string or a list of strings.")


def is_base64_encoded(s: str) -> bool:
    """
    Check if a string is Base64 encoded.
    :param s: The string to check.
    :return: True if the string is Base64 encoded, False otherwise.
    """
    # Base64 encoded string length should be a multiple of 4, and only contain Base64 character set
    if len(s) % 4 != 0 or not re.match(r"^[A-Za-z0-9+/]*={0,2}$", s):
        return False

    try:
        # Try to decode the string
        base64.b64decode(s, validate=True)
        return True
    except Exception:
        return False


def image_to_base64(
    image: Image.Image | str,
    quality: int = 85,
    max_resolution: tuple = (448, 448),
    img_format: str = "WEBP",
) -> str:
    """
    Convert the image to a base64 encoded string.

    :param image: PIL Image object or the path to the image file.
    :param quality: Compression quality (0-100) for WebP format. Higher values mean better quality but larger size.
    :param max_resolution: Optional maximum resolution (width, height) to fit the image within while preserving aspect ratio.
    :param img_format: Image format to use for encoding. Default is "WEBP".
    :return: Base64 encoded string of the image.
    """
    if isinstance(image, str):
        # If the image is a string, assume it's a path and open it
        image = Image.open(image)

    # Convert the image to RGB mode if it's in RGBA mode
    if image.mode == "RGBA":
        image = image.convert("RGB")

    # Resize the image while preserving aspect ratio
    if max_resolution:
        original_width, original_height = image.size
        max_width, max_height = max_resolution

        # Calculate the new dimensions while preserving aspect ratio
        ratio = min(max_width / original_width, max_height / original_height)
        new_width = int(original_width * ratio)
        new_height = int(original_height * ratio)

        # Resize the image
        image = image.resize((new_width, new_height), Image.Resampling.LANCZOS)

    # Remove metadata (EXIF, etc.)
    if "exif" in image.info:
        del image.info["exif"]

    # Save the image to a BytesIO buffer as WebP with specified quality
    buffered = io.BytesIO()
    image.save(buffered, format=img_format, quality=quality)

    # Return the Base64 encoded string
    return base64.b64encode(buffered.getvalue()).decode("utf-8")


def find_files_with_extensions(
    directory_path: str,
    extensions: str | Iterable[str],
    return_all: bool = False,
) -> str | List[str] | None:
    """
    Find files with the specified extensions in the given directory.
    If `return_all` is False (default), returns the first matching file based on priority.
    If `return_all` is True, returns a list of all matching files, sorted by priority.

    Args:
        directory_path (str): Path to the directory to search.
        extensions (Union[str, List[str]]): A single extension or a list of extensions.
        return_all (bool): If True, return all matching files; otherwise, return the first match.

    Returns:
        Union[str, List[str], None]: A single file path, a list of file paths, or None if no files are found.
    """

    def _validate_directory_path(path: str):
        """Validate directory path exists and is accessible"""
        if not os.path.exists(path):
            raise FileNotFoundError(f"Path '{path}' does not exist.")
        if not os.path.isdir(path):
            raise ValueError(f"Path '{path}' is not a directory.")

    def _find_matching_files(directory_path: str, extensions: List[str]) -> List[str]:
        """Find all files matching extensions in directory"""
        matching_files = []
        for root, _, files in os.walk(directory_path):
            for file in files:
                if any(file.endswith(f".{ext}") for ext in extensions):
                    matching_files.append(os.path.join(root, file))
        return matching_files

    def _sort_files_by_priority(files: List[str], extensions: List[str]) -> List[str]:
        """Sort files by extension priority"""
        return sorted(files, key=lambda f: extensions.index(os.path.splitext(f)[1][1:]))

    try:
        extensions = [extensions] if isinstance(extensions, str) else extensions
        _validate_directory_path(directory_path)

        matching_files = _find_matching_files(directory_path, extensions)

        if not matching_files:
            return None if not return_all else []

        if return_all:
            return _sort_files_by_priority(matching_files, extensions)

        return matching_files[0]

    except (FileNotFoundError, PermissionError) as e:
        logger.error(str(e))
        return None if not return_all else []


def extract_section_markdown(text: str, heading: str) -> str:
    """
    Extracts content from a markdown text under the specified heading.
    """
    lines = text.split("\n")
    content = []
    capture = False
    for line in lines:
        line = line.strip()
        if line.startswith(f"#{heading}"):
            capture = True
            continue
        if line.startswith("#"):
            capture = False
            continue
        if capture:
            content.append(line)
    return "\n".join(content).strip()


def parse_json_response(response: str) -> Dict[str, Any]:
    """Parse JSON response from VLM, handling potential errors"""
    try:
        # Normalize bracket usage for JSON parsing
        response = response.replace("{{", "{").replace("}}", "}")

        # Extract JSON content from response
        json_start = response.find("{")
        json_end = response.rfind("}") + 1
        json_str = response[json_start:json_end]

        return json.loads(json_str)
    except json.JSONDecodeError as e:
        logger.error(f"Failed to parse JSON response: {str(e)}")
        logger.debug(f"Original response: {response}")
        return {}
    except Exception as e:
        logger.error(f"Unexpected error parsing response: {str(e)}")
        return {}


def parse_design_goal(design_goal_input: str) -> str:
    """
    Parse the design goal input, which can be either a JSON file or plain text.
    If it's a JSON file, extract the 'text' field.

    Args:
        design_goal_input (str): Path to a JSON file or plain text.

    Returns:
        str: The design goal text.
    """
    if os.path.isfile(design_goal_input):
        with open(design_goal_input, "r") as f:
            try:
                data = json.load(f)
                return data.get("text", "")
            except json.JSONDecodeError:
                logger.error("The provided file is not a valid JSON.")
                raise json.JSONDecodeError("The provided file is not a valid JSON.")
    return design_goal_input


class ANSIStrippingFormatter(logging.Formatter):
    """
    A logging formatter that removes ANSI escape codes from log messages.

    This formatter is particularly useful when logging to files, where ANSI escape codes
    (used for terminal coloring) are not needed and can clutter the log output.
    """

    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

    def format(self, record):
        """
        Format the log record, removing any ANSI escape codes from the message.

        Args:
            record (logging.LogRecord): The log record to format.

        Returns:
            str: The formatted log message with ANSI escape codes removed.
        """
        message = super().format(record)
        return self.ansi_escape.sub("", message)


def setup_logging(
    log_level: str = "INFO",
    log_file: Optional[str] = None,
    log_format: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
) -> None:
    """
    Configure the logging system.

    This function sets up logging with the following behavior:
    - Console logs retain ANSI escape codes for colored output.
    - File logs (if specified) have ANSI escape codes removed for cleaner output.
    - Mutes httpx INFO logs while keeping the global log_level at INFO.

    Args:
        log_level (str): The logging level (e.g., "INFO", "DEBUG"). Defaults to "INFO".
        log_file (Optional[str]): Path to the log file. If None, no file logging is performed.
        log_format (str): The format string for log messages. Defaults to "%(asctime)s - %(name)s - %(levelname)s - %(message)s".

    Returns:
        None
    """
    # Define formatters
    formatters = {
        "console": {
            "format": log_format,
        },
        "file": {
            "()": ANSIStrippingFormatter,  # Remove ANSI escape codes for file logs
            "format": log_format,
        },
    }

    # define handlers (default enable console)
    handlers = {
        "console": {
            "level": log_level.upper(),
            "class": "logging.StreamHandler",
            "formatter": "console",
            "stream": "ext://sys.stdout",
        }
    }

    # only add file handler when log_file is provided
    if log_file:
        handlers["file"] = {
            "level": log_level.upper(),
            "class": "logging.FileHandler",
            "filename": log_file,
            "formatter": "file",
            "mode": "w",
        }

    # configure logging
    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": formatters,
        "handlers": handlers,
        "loggers": {
            "": {
                "level": log_level.upper(),
                "handlers": list(handlers.keys()),
                "propagate": True,  # Logs propagate to root logger
            },
            # Add a custom logger for httpx to mute INFO logs
            "httpx": {
                "level": "WARNING",  # Set httpx logger to WARNING to mute INFO logs
                "handlers": list(handlers.keys()),
                "propagate": False,  # Prevent httpx logs from propagating to root logger
            },
        },
    }

    logging.config.dictConfig(logging_config)
    logging.info("Logging configuration is set up.")


def recover_stream_tool_calls(
    stream_tool_calls: Dict,
) -> List[ChatCompletionMessageToolCall]:
    tool_calls = []
    for index, tool_call_str in stream_tool_calls.items():
        tool_call_instance = ChatCompletionMessageToolCall(
            id=tool_call_str["id"],
            type="function",
            function=Function(
                name=tool_call_str["function"]["name"],
                arguments=tool_call_str["function"]["arguments"],
            ),
            index=index,
        )
        tool_calls.append(tool_call_instance)
    return tool_calls


if __name__ == "__main__":
    cprint("This is a red message", "red")
    cprint("This is a green message", "green")
    cprint("This is a blue message", "blue")
    cprint("This is a yellow message", "yellow")
    cprint("This is a magenta message", "magenta")
    cprint("This is a cyan message", "cyan")
    cprint("This is a black message", "black")
    cprint("This is a white message", "white")

    # Configure logging
    setup_logging()

    # Example usage:
    logger.info(colorstring("This is a red message", "red"))
    logger.info(colorstring("This is a green message", "green"))
    logger.info(colorstring("This is a blue message", "blue"))
    logger.info(colorstring("This is a yellow message", "yellow"))
    logger.info(colorstring("This is a magenta message", "magenta"))
    logger.info(colorstring("This is a cyan message", "cyan"))
    logger.info(colorstring("This is a black message", "black"))
    logger.info(colorstring("This is a white message", "white"))



import logging
import signal
import sys
from typing import Any, Dict, Tuple

import questionary
from questionary import Style

from cicada.core import model
from cicada.core.basics import DesignGoal, PromptBuilder
from cicada.core.utils import colorstring, parse_json_response


logger = logging.getLogger(__name__)

MAX_IMAGES_PER_QUERY = 4  # to prevent input exceed max input token

# Define a custom style for questionary prompts
custom_style = Style(
    [
        ("question", "fg:#ff0000 bold"),  # Red and bold for questions
        ("answer", "fg:#00ff00 underline"),  # Green and underlined for answers
    ]
)


class Describer(model.MultiModalModel):
    def __init__(
        self,
        api_key,
        api_base_url,
        model_name,
        org_id,
        prompt_templates,
        **model_kwargs,
    ):
        super().__init__(
            api_key,
            api_base_url,
            model_name,
            org_id,
            **model_kwargs,
        )
        self.reverse_engineer_prompt = prompt_templates.get("reverse_engineer", {})
        self.featurize_design_prompt = prompt_templates.get("featurize_design", {})

    def featurize_design_goal_with_confidence(
        self, design_goal: DesignGoal, user_feedback: str = None
    ) -> Tuple[Dict[str, Any], str]:
        """
        Generate a refined design based on the user's goal.
        Returns a tuple of (parsed JSON result, raw response).
        """
        text_goal = design_goal.text
        ref_images = design_goal.images
        logging.debug(
            f"Featurizing design goal: {text_goal}, with images: {ref_images}"
        )

        pb = PromptBuilder()
        pb.add_system_message(self.featurize_design_prompt["system_prompt_template"])

        pb.add_user_message(f"The current design goal: '{text_goal}'.")

        if ref_images:
            pb.add_text("The following reference images are provided:")
            pb.add_images(ref_images)

        if user_feedback:
            pb.add_text(
                "The user has provided the following feedback, please revise the current design against the feedback:"
            )
            pb.add_text(user_feedback)

        pb.add_user_message(self.featurize_design_prompt["user_prompt_template"])

        # Query the VLM with images and prompt
        response = self.query(prompt_builder=pb, stream=self.stream)["content"]

        # Parse the JSON response
        json_result = parse_json_response(response)

        return json_result, response

    def decompose_design(self, design_goal: DesignGoal) -> Tuple[Dict[str, Any], str]:
        """
        Decompose the design into its constituent parts, specifying building methods and geometric details.
        Returns a tuple of (parsed JSON result, raw response).
        """
        text_goal = design_goal.text
        ref_images = design_goal.images
        logging.debug(
            f"Decomposing design goal: {text_goal}, with images: {ref_images}"
        )

        # Dynamically construct the text_goal_section based on whether text_goal is provided
        if text_goal:
            text_goal_section = (
                f"The user has provided the following design goal: '{text_goal}'. "
                "Use this as the primary input for reverse engineering the design. "
                "The provided images are for reference only.\n\n"
            )
        else:
            text_goal_section = (
                "The user has provided the following images. "
                "Use these as the primary input for reverse engineering the design.\n\n"
            )

        # Replace the {text_goal_section} placeholder in the user prompt template
        user_prompt = self.reverse_engineer_prompt["user_prompt_template"].format(
            text_goal_section=text_goal_section
        )

        # Prepare the prompt for decomposing the design
        pb = PromptBuilder()
        pb.add_system_message(self.reverse_engineer_prompt["system_prompt_template"])
        pb.add_user_message(user_prompt)

        if ref_images:
            pb.add_text("The following is a set of reference images:")
            pb.add_images(ref_images)

        # Query the VLM with images and prompt
        response = self.query(prompt_builder=pb, stream=self.stream)["content"]

        # Parse the JSON response
        json_result = parse_json_response(response)

        return json_result, response

    def _analyze_text_goal_against_images(
        self, text_goal: str, ref_images: list
    ) -> bool:
        """
        Analyze the text goal against reference images and ask for confirmation if there's a conflict.
        Returns True if the user confirms to proceed, False otherwise.
        """
        logging.debug(f"Analyzing text goal against reference images: {text_goal}")

        # Prepare a prompt to analyze the text goal against images
        analysis_prompt = (
            "The user has provided the following text goal:\n"
            f"{text_goal}\n\n"
            "The following reference images are also provided:\n"
            f"{ref_images}\n\n"
            "Your task is to analyze whether the text goal aligns with the reference images. "
            "If there is a conflict, highlight it and ask for user confirmation.\n\n"
            "Return the analysis in JSON format."
        )

        pb = PromptBuilder()
        pb.add_system_message(self.featurize_design_prompt["system_prompt_template"])
        pb.add_user_message(analysis_prompt)

        # Query the VLM with the analysis prompt
        response = self.query(prompt_builder=pb, stream=self.stream)["content"]
        analysis_result = parse_json_response(response)

        # If there's a conflict, ask for user confirmation
        if analysis_result.get("conflict_detected", False):
            print(
                colorstring(
                    "\nConflict detected between text goal and reference images.",
                    "bright_red",
                )
            )
            confirmation = questionary.confirm(
                "The system detected a conflict between your text goal and reference images. Do you want to proceed with the current text goal?",
                style=custom_style,
            ).ask()

            if not confirmation:
                print(
                    colorstring(
                        "Please revise your text goal or reference images and try again.",
                        "bright_yellow",
                    )
                )
                return False

        return True

    def design_feedback_loop(
        self,
        design: DesignGoal,
        iteration: int = 1,
        max_iterations: int = 5,
        feedback: str = None,
    ) -> DesignGoal:
        """
        Execute the complete design feedback cycle using recursion.
        """

        # Handle Ctrl+C gracefully
        def signal_handler(sig, frame):
            print("\nProcess interrupted by user. Exiting...")
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)

        # Base case: Maximum iterations reached
        if iteration > max_iterations:
            logger.warning(f"Maximum iterations ({max_iterations}) reached")
            # decompose the design and return the result
            updated_design = self._update_design_with_decomposition(design)

            return updated_design

        logger.info(f"Starting iteration {iteration}/{max_iterations}")
        try:
            logging.debug(f"Current Design: {design.text}")

            # At the start of the first iteration, check for conflicts between text goal and reference images
            if iteration == 1 and design.images:
                if not self._analyze_text_goal_against_images(
                    design.text, design.images
                ):
                    sys.exit(0)

            # Generate VLM response and parse the JSON result
            updated_design_result, response = (
                self.featurize_design_goal_with_confidence(design, feedback)
            )
            updated_design_text = updated_design_result.get(
                "current_design", design.text
            )
            updated_design = DesignGoal(updated_design_text, design.images)

            # Display the proposed design to the user
            print(f"\n{'='*40}\nIteration {iteration}/{max_iterations}")
            print(f"Proposed Design:\n{updated_design_text}")

            # Always require confirmation after the first iteration
            if iteration == 1:
                print(
                    colorstring(
                        "\nThe system requires your feedback or confirmation after the first iteration.",
                        "bright_yellow",
                    )
                )
            elif updated_design_result.get("needs_human_validation", True):
                print(
                    colorstring(
                        "\nThe system requires your feedback or confirmation.",
                        "bright_yellow",
                    )
                )
            else:
                print(
                    colorstring(
                        "\nThe system is confident in this solution.", "bright_blue"
                    )
                )

            # Always allow the user to provide feedback            # Define static choices once
            choices = [
                {"name": "Provide Feedback", "value": "feedback"},
                {"name": "Confirm", "value": "confirm"},
                {"name": "Exit", "value": "exit"},
            ]
            # Multi-choice prompt with dynamic choices
            action = questionary.select(
                "What would you like to do?",
                choices=choices,
                style=custom_style,
            ).ask()

            if action == "exit":
                logger.info("Exiting feedback loop by user request")
                return updated_design

            elif action == "confirm":
                # Confirm the design
                logger.info("Design confirmed. Decomposing the final design...")
                updated_design = self._update_design_with_decomposition(updated_design)

                return updated_design

            elif action == "feedback":
                # Allow the user to provide custom feedback
                feedback = questionary.text(
                    "Please provide your feedback:",
                    validate=lambda text: len(text.strip()) > 0
                    or "Feedback cannot be empty.",
                    style=custom_style,
                ).ask()

                # Recursively call the function with the updated design, incremented iteration, and feedback
                return self.design_feedback_loop(
                    updated_design, iteration + 1, max_iterations, feedback
                )

        except Exception as e:
            logger.error(f"Iteration {iteration} failed: {e}")
            raise e

    def _update_design_with_decomposition(self, updated_design) -> DesignGoal:
        decomposition_result, _ = self.decompose_design(updated_design)
        logger.info(f"Decomposition Result:\n{decomposition_result}")
        print(f"\n{'='*40}\nDecomposition Result:\n{decomposition_result}")

        # Store the decomposition result in design_goal.extra
        if not hasattr(updated_design, "extra"):
            updated_design.extra = {}
        updated_design.extra["decomposition"] = decomposition_result
        return updated_design


if __name__ == "__main__":
    import argparse
    import json
    import os

    from cicada.core.utils import (
        colorstring,
        load_config,
        load_prompts,
        setup_logging,
    )

    setup_logging()

    parser = argparse.ArgumentParser(description="Vision Language Model")
    parser.add_argument(
        "--config", default="config.yaml", help="Path to the configuration YAML file"
    )
    parser.add_argument(
        "--prompts", default="prompts/", help="Path to the prompts YAML file"
    )
    parser.add_argument(
        "text_goal",
        type=str,
        help="The text goal for the design",
    )
    parser.add_argument(
        "-img",
        "--ref_images",
        type=str,
        default=None,
        help="Paths to reference images for the design",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="/tmp/cicada/refined_design_goal.json",
        help="Path to the output JSON file",
    )
    args = parser.parse_args()

    describe_vlm_config = load_config(args.config, "describe-vlm")

    vlm = Describer(
        describe_vlm_config["api_key"],
        describe_vlm_config.get("api_base_url"),
        describe_vlm_config.get("model_name", "gpt-4o"),
        describe_vlm_config.get("org_id"),
        load_prompts(args.prompts, "describe-vlm"),
        **describe_vlm_config.get("model_kwargs", {}),
    )

    design_goal = DesignGoal(args.text_goal, args.ref_images)
    print(design_goal)

    # Run the feedback loop process
    try:
        refined_design = vlm.design_feedback_loop(design_goal)
        logger.info("Design process completed successfully")
        logger.info(colorstring(f"Refined Design:\n{refined_design}", "white"))

        if args.output:
            # make outdir
            os.makedirs(os.path.dirname(args.output), exist_ok=True)
            # save refined design to output file
            with open(args.output, "w") as f:
                json.dump(refined_design.to_dict(), f, indent=4)
            logger.info(f"Refined design saved to {args.output}")

    except KeyboardInterrupt:
        logger.warning("Process interrupted by user")
        sys.exit(1)
    except Exception as e:
        logger.error(f"Design process failed: {e}")
        sys.exit(1)



import logging
import os
from typing import Tuple

import numpy as np
import trimesh
from plyfile import PlyData, PlyElement

# TODO: should be adaptive to the size of the mesh
POINTCLOUD_N_POINTS = 8096 * 3
DEFAULT_RGB = (88, 88, 88)


logger = logging.getLogger(__name__)


def recenter_and_reaxis_mesh(
    mesh: trimesh.Trimesh,
) -> Tuple[trimesh.Trimesh, np.ndarray]:
    """
    Recenter and reaxis the mesh using its principal inertia transform,
    ensuring that the mesh is oriented such that the gravity vector (negative Z-axis)
    aligns with the principal axis corresponding to the object's "up" direction.

    Args:
        mesh (trimesh.Trimesh): The input mesh.

    Returns:
        Tuple[trimesh.Trimesh, np.ndarray]: The transformed mesh and the transformation matrix.
    """
    # Get the principal inertia transform
    transformation_matrix = mesh.principal_inertia_transform
    logger.debug(f"[Transform] principal inertia transform: \n{transformation_matrix}")

    # Apply the transformation to the mesh
    transformed_mesh = mesh.copy()
    transformed_mesh.apply_transform(transformation_matrix)

    # Ensure that the mesh is oriented such that the gravity vector (negative Z-axis)
    # aligns with the principal axis corresponding to the object's "up" direction.
    # We assume that the principal axis corresponding to the smallest moment of inertia
    # is the "up" direction.
    moments_of_inertia = transformed_mesh.moment_inertia
    principal_axes = np.diag(moments_of_inertia)

    # Find the index of the smallest moment of inertia (assumed to be the "up" direction)
    up_axis_index = np.argmin(principal_axes)

    # Adjusted rotation logic to align the object's "up" direction with +Y (trimesh's upward direction)
    # 调整旋转逻辑，将物体的“上”方向对齐到+Y轴（trimesh的垂直正方向）
    if up_axis_index == 0:
        # X-axis -> Rotate 90° around Z to align X to Y
        rotation_matrix = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    elif up_axis_index == 2:
        # Z-axis -> Rotate -90° around X to align Z to Y
        rotation_matrix = np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]])
    else:
        # Y-axis (no rotation needed)
        rotation_matrix = np.eye(3)

    # Convert the 3x3 rotation matrix to a 4x4 transformation matrix
    rotation_matrix_4x4 = np.eye(4)
    rotation_matrix_4x4[:3, :3] = rotation_matrix

    # Apply the rotation to the transformed mesh
    transformed_mesh.apply_transform(rotation_matrix_4x4)

    # Combine the original transformation matrix with the rotation matrix
    combined_transform = np.dot(rotation_matrix_4x4, transformation_matrix)
    transformation_matrix = combined_transform

    logger.debug(
        f"[Transform] mesh transformed with principal inertia transform and aligned with gravity"
    )
    return transformed_mesh, transformation_matrix


def step2mesh(step_path: str) -> trimesh.Trimesh:
    """
    Convert a STEP file to a mesh using GMSH.

    Args:
        step_path (str): Path to the STEP file.

    Returns:
        trimesh.Trimesh: The generated mesh.
    """
    mesh = trimesh.Trimesh(
        **trimesh.interfaces.gmsh.load_gmsh(
            file_name=step_path,
            gmsh_args=[
                ("Mesh.Algorithm", 5),
                ("Mesh.Algorithm3D", 1),
                ("Mesh.CharacteristicLengthFromCurvature", 50),
                ("General.NumThreads", 10),
                ("Mesh.MinimumCirclePoints", 32),
            ],
        )
    )
    logger.debug(
        f"[step2mesh] mesh loaded from {step_path}, {mesh.vertices.shape[0]} vertices"
    )
    return mesh


def step2obj(step_path: str, out_path: str) -> str:
    """
    Convert a STEP file to an OBJ file.

    Args:
        step_path (str): Path to the STEP file.
        out_path (str): Directory to save the OBJ file.

    Returns:
        str: Path to the saved OBJ file.
    """
    m = step2mesh(step_path)
    m, _ = recenter_and_reaxis_mesh(m)

    obj_path = os.path.join(
        out_path, os.path.basename(step_path).replace(".step", ".obj")
    )
    m.export(obj_path, file_type="obj")
    logger.info(f"[step2obj] mesh saved to {obj_path}")

    return obj_path


def step2stl(step_path: str, out_path: str) -> str:
    """
    Convert a STEP file to an STL file.

    Args:
        step_path (str): Path to the STEP file.
        out_path (str): Directory to save the STL file.

    Returns:
        str: Path to the saved STL file.
    """
    mesh = step2mesh(step_path)
    stl_path = os.path.join(
        out_path, os.path.basename(step_path).replace(".step", ".stl")
    )
    mesh.export(stl_path, file_type="stl")
    logger.info(f"STEP file converted to STL: {stl_path}")

    return stl_path


def stl2obj(stl_path: str, out_path: str) -> str:
    """
    Convert an STL file to an OBJ file.

    Args:
        stl_path (str): Path to the STL file.
        out_path (str): Directory to save the OBJ file.

    Returns:
        str: Path to the saved OBJ file.
    """
    mesh = trimesh.load_mesh(stl_path)
    obj_path = os.path.join(
        out_path, os.path.basename(stl_path).replace(".stl", ".obj")
    )
    mesh.export(obj_path, file_type="obj")
    logger.info(f"STL file converted to OBJ: {obj_path}")

    return obj_path


def obj2stl(obj_path: str, out_path: str) -> str:
    """
    Convert an OBJ file to an STL file.

    Args:
        obj_path (str): Path to the OBJ file.
        out_path (str): Directory to save the STL file.

    Returns:
        str: Path to the saved STL file.
    """
    mesh = trimesh.load_mesh(obj_path)
    stl_path = os.path.join(
        out_path, os.path.basename(obj_path).replace(".obj", ".stl")
    )
    mesh.export(stl_path, file_type="stl")
    logger.info(f"OBJ file converted to STL: {stl_path}")

    return stl_path


def write_ply(
    points: np.ndarray,
    filename: str,
    text: bool = False,
    default_rgb: Tuple[int, int, int] = DEFAULT_RGB,
) -> None:
    """
    Write points to a PLY file.

    Args:
        points (np.ndarray): Nx3 array of points.
        filename (str): Path to save the PLY file.
        text (bool): Whether to save the PLY file in text format.
        default_rgb (Tuple[int, int, int]): Default RGB color for the points.
    """
    points = [
        (
            points[i, 0],
            points[i, 1],
            points[i, 2],
            default_rgb[0],
            default_rgb[1],
            default_rgb[2],
        )
        for i in range(points.shape[0])
    ]
    vertex = np.array(
        points,
        dtype=[
            ("x", "f4"),
            ("y", "f4"),
            ("z", "f4"),
            ("red", "u1"),
            ("green", "u1"),
            ("blue", "u1"),
        ],
    )
    el = PlyElement.describe(vertex, "vertex", comments=["vertices"])
    with open(filename, mode="wb") as f:
        PlyData([el], text=text).write(f)
    logger.debug(f"[write_ply] saved to {filename}, with default rgb is {default_rgb}")


def obj2pc(obj_path: str, out_path: str) -> str:
    """
    Convert an OBJ file to a point cloud (PLY format).

    Args:
        obj_path (str): Path to the OBJ file.
        out_path (str): Directory to save the PLY file.

    Returns:
        str: Path to the saved PLY file.
    """
    m = trimesh.load_mesh(obj_path)
    logger.debug(
        f"[obj2pc] mesh loaded from {obj_path} with {m.vertices.shape[0]} vertices"
    )
    pc_path = os.path.join(out_path, os.path.basename(obj_path).replace(".obj", ".ply"))
    pc = trimesh.PointCloud(m.sample(POINTCLOUD_N_POINTS))
    logger.debug(f"[obj2pc] convert to pointcloud, with {pc.vertices.shape[0]} points")

    pc = pc.vertices
    write_ply(pc, pc_path)
    logger.info(f"[obj2pc] pointcloud saved to {pc_path}")

    return pc_path


def stl2pc(stl_path: str, out_path: str) -> str:
    """
    Convert an STL file to a point cloud (PLY format).

    Args:
        stl_path (str): Path to the STL file.
        out_path (str): Directory to save the PLY file.

    Returns:
        str: Path to the saved PLY file.
    """
    mesh = trimesh.load_mesh(stl_path)
    pc_path = os.path.join(out_path, os.path.basename(stl_path).replace(".stl", ".ply"))
    pc = trimesh.PointCloud(mesh.sample(POINTCLOUD_N_POINTS))
    write_ply(pc.vertices, pc_path)
    logger.info(f"STL file converted to point cloud: {pc_path}")

    return pc_path


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--obj_file", type=str)
    group.add_argument("--step_file", type=str)
    group.add_argument("--stl_file", type=str)

    group_action = parser.add_mutually_exclusive_group(required=True)
    group_action.add_argument("--convert_step2obj", action="store_true")
    group_action.add_argument("--convert_obj2pc", action="store_true")
    group_action.add_argument("--convert_step2stl", action="store_true")
    group_action.add_argument("--convert_obj2stl", action="store_true")
    group_action.add_argument("--convert_stl2obj", action="store_true")
    group_action.add_argument("--convert_stl2pc", action="store_true")
    group_action.add_argument(
        "--reaxis_gravity",
        action="store_true",
        help="Recenter and reaxis the mesh to align with gravity.",
    )

    args = parser.parse_args()

    if args.obj_file:
        obj_file = args.obj_file
        out_path = os.path.dirname(obj_file)
        if args.convert_obj2pc:
            obj2pc(obj_file, out_path)
        elif args.convert_obj2stl:
            obj2stl(obj_file, out_path)
        elif args.reaxis_gravity:
            mesh = trimesh.load_mesh(obj_file)
            transformed_mesh, _ = recenter_and_reaxis_mesh(mesh)
            transformed_mesh.export(obj_file, file_type="obj")
            logger.info(f"Mesh reoriented with gravity and saved to {obj_file}")
        else:
            logger.error("No valid action selected for OBJ file.")
    elif args.step_file:
        step_file = args.step_file
        out_path = os.path.dirname(step_file)
        if args.convert_step2obj:
            step2obj(step_file, out_path)
        elif args.convert_step2stl:
            step2stl(step_file, out_path)
        else:
            logger.error("No valid action selected for STEP file.")
    elif args.stl_file:
        stl_file = args.stl_file
        out_path = os.path.dirname(stl_file)
        if args.convert_stl2obj:
            stl2obj(stl_file, out_path)
        elif args.convert_stl2pc:
            stl2pc(stl_file, out_path)
        elif args.reaxis_gravity:
            mesh = trimesh.load_mesh(stl_file)
            transformed_mesh, _ = recenter_and_reaxis_mesh(mesh)
            transformed_mesh.export(stl_file, file_type="stl")
            logger.info(f"Mesh reoriented with gravity and saved to {stl_file}")
        else:
            logger.error("No valid action selected for STL file.")
